# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Tests for the MindtPy solver."""

import pyomo.common.unittest as unittest
from pyomo.contrib.mindtpy.tests.eight_process_problem import EightProcessFlowsheet
from pyomo.contrib.mindtpy.tests.MINLP_simple import SimpleMINLP as SimpleMINLP
from pyomo.contrib.mindtpy.tests.MINLP2_simple import SimpleMINLP as SimpleMINLP2
from pyomo.contrib.mindtpy.tests.MINLP3_simple import SimpleMINLP as SimpleMINLP3
from pyomo.contrib.mindtpy.tests.from_proposal import ProposalModel
from pyomo.contrib.mindtpy.tests.constraint_qualification_example import (
    ConstraintQualificationExample,
)
from pyomo.environ import SolverFactory, value
from pyomo.opt import TerminationCondition

if SolverFactory('appsi_highs').available(exception_flag=False) and SolverFactory(
    'appsi_highs'
).version() >= (1, 7, 0):
    required_solvers = ('ipopt', 'appsi_highs')
else:
    required_solvers = ('ipopt', 'glpk')

if all(SolverFactory(s).available(exception_flag=False) for s in required_solvers):
    subsolvers_available = True
else:
    subsolvers_available = False

model_list = [
    EightProcessFlowsheet(convex=True),
    ConstraintQualificationExample(),
    SimpleMINLP(),
    SimpleMINLP2(),
    SimpleMINLP3(),
    ProposalModel(),
]


@unittest.skipIf(
    not subsolvers_available,
    'Required subsolvers %s are not available' % (required_solvers,),
)
class TestMindtPy(unittest.TestCase):
    """Tests for the MindtPy solver plugin."""

    def check_optimal_solution(self, model, places=1):
        for var in model.optimal_solution:
            self.assertAlmostEqual(
                var.value, model.optimal_solution[var], places=places
            )

    def test_ECP(self):
        """Test the extended cutting plane decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='ECP',
                    init_strategy='rNLP',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    absolute_bound_tolerance=1e-5,
                )

                self.assertIs(
                    results.solver.termination_condition, TerminationCondition.optimal
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)

    def test_ECP_add_slack(self):
        """Test the extended cutting plane decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='ECP',
                    init_strategy='rNLP',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    absolute_bound_tolerance=1e-5,
                    add_slack=True,
                )

                self.assertIs(
                    results.solver.termination_condition, TerminationCondition.optimal
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=1
                )
                self.check_optimal_solution(model)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Transformation to detect variables fixed by bounds and fix them."""

from math import fabs

from pyomo.core.base.transformation import TransformationFactory
from pyomo.common.collections import ComponentMap
from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    NonNegativeFloat,
    document_kwargs_from_configdict,
)
from pyomo.core.base.var import Var
from pyomo.core.expr.numvalue import value
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.core.base.block import Block
from pyomo.gdp import Disjunct


@TransformationFactory.register(
    'contrib.detect_fixed_vars',
    doc="Detect variables that are de-facto fixed but not considered fixed.",
)
@document_kwargs_from_configdict('CONFIG')
class FixedVarDetector(IsomorphicTransformation):
    """Detects variables that are de-facto fixed but not considered fixed.

    For each variable :math:`v` found on the model, check to see if its lower
    bound :math:`v^{LB}` is within some tolerance of its upper bound
    :math:`v^{UB}`. If so, fix the variable to the value of :math:`v^{LB}`.

    Keyword arguments below are specified for the ``apply_to`` and
    ``create_using`` functions.

    """

    CONFIG = ConfigBlock("FixedVarDetector")
    CONFIG.declare(
        "tmp",
        ConfigValue(
            default=False,
            domain=bool,
            description="True to store the set of transformed variables and "
            "their old values so that they can be restored.",
        ),
    )
    CONFIG.declare(
        "tolerance",
        ConfigValue(
            default=1e-13,
            domain=NonNegativeFloat,
            description="tolerance on bound equality (LB == UB)",
        ),
    )

    def _apply_to(self, instance, **kwargs):
        config = self.CONFIG(kwargs)

        if config.tmp:
            instance._xfrm_detect_fixed_vars_old_values = ComponentMap()

        for var in instance.component_data_objects(
            ctype=Var, descend_into=[Block, Disjunct]
        ):
            if var.fixed or var.lb is None or var.ub is None:
                # if the variable is already fixed, or if it is missing a
                # bound, we skip it.
                continue
            if fabs(value(var.lb) - value(var.ub)) <= config.tolerance:
                if config.tmp:
                    instance._xfrm_detect_fixed_vars_old_values[var] = var.value
                var.fix(var.lb)

    def revert(self, instance):
        """Revert variables fixed by the transformation."""
        for var, var_value in instance._xfrm_detect_fixed_vars_old_values.items():
            var.unfix()
            var.set_value(var_value)

        del instance._xfrm_detect_fixed_vars_old_values



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from pyomo.opt.base.error import ConverterError
from pyomo.opt.base.convert import convert_problem
from pyomo.opt.base.solvers import (
    UnknownSolver,
    SolverFactory,
    check_available_solvers,
    OptSolver,
)
from pyomo.opt.base.results import ReaderFactory, AbstractResultsReader
from pyomo.opt.base.problem import AbstractProblemWriter, BranchDirection, WriterFactory
from pyomo.opt.base.formats import ProblemFormat, ResultsFormat, guess_format



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import inspect
import logging
import sys
from copy import deepcopy
from collections import deque

logger = logging.getLogger('pyomo.core')

from pyomo.common.deprecation import deprecated, deprecation_warning
from pyomo.common.errors import DeveloperError, TemplateExpressionError
from pyomo.common.numeric_types import (
    nonpyomo_leaf_types,
    native_types,
    native_numeric_types,
    value,
)
import pyomo.core.expr.expr_common as common
from pyomo.core.expr.symbol_map import SymbolMap

try:
    # sys._getframe is slightly faster than inspect's currentframe, but
    # is not guaranteed to exist everywhere
    currentframe = sys._getframe
except AttributeError:
    currentframe = inspect.currentframe


def get_stack_depth():
    n = -1  # skip *this* frame in the count
    f = currentframe()
    while f is not None:
        n += 1
        f = f.f_back
    return n


# For efficiency, we want to run recursively, but don't want to hit
# Python's recursion limit (because that would be difficult to recover
# from cleanly).  However, there is a non-trivial cost to determine the
# current stack depth - and we don't want to hit that for every call.
# Instead, we will assume that the walker is always called with at
# least RECURSION_LIMIT frames available on the stack.  When we hit the
# end of that limit, we will actually check how much space is left on
# the stack and run recursively until only 2*RECURSION_LIMIT frames are
# left.  For the vast majority of well-formed expressions this approach
# avoids a somewhat costly call to get_stack_depth, but still catches
# the vast majority of cases that could generate a recursion error.
RECURSION_LIMIT = 50


class RevertToNonrecursive(Exception):
    pass


# NOTE: This module also has dependencies on numeric_expr; however, to
# avoid circular dependencies, we will NOT import them here.  Instead,
# until we can resolve the circular dependencies, they will be injected
# into this module by the .current module (which must be imported
# *after* numeric_expr, logocal_expr, and this module.


# -------------------------------------------------------
#
# Visitor Logic
#
# -------------------------------------------------------


class StreamBasedExpressionVisitor:
    """This class implements a generic stream-based expression walker.

    This visitor walks an expression tree using a depth-first strategy
    and generates a full event stream similar to other tree visitors
    (e.g., the expat XML parser).  The following events are triggered
    through callback functions as the traversal enters and leaves nodes
    in the tree:

    ::

       initializeWalker(expr) -> walk, result
       enterNode(N1) -> args, data
       {for N2 in args:}
         beforeChild(N1, N2) -> descend, child_result
           enterNode(N2) -> N2_args, N2_data
           [...]
           exitNode(N2, n2_data) -> child_result
         acceptChildResult(N1, data, child_result) -> data
         afterChild(N1, N2) -> None
       exitNode(N1, data) -> N1_result
       finalizeWalker(result) -> result

    Individual event callbacks match the following signatures:

    walk, result = initializeWalker(self, expr):

         initializeWalker() is called to set the walker up and perform
         any preliminary processing on the root node.  The method returns
         a flag indicating if the tree should be walked and a result.  If
         `walk` is True, then result is ignored.  If `walk` is False,
         then `result` is returned as the final result from the walker,
         bypassing all other callbacks (including finalizeResult).

    args, data = enterNode(self, node):

         enterNode() is called when the walker first enters a node (from
         above), and is passed the node being entered.  It is expected to
         return a tuple of child `args` (as either a tuple or list) and a
         user-specified data structure for collecting results.  If None
         is returned for args, the node's args attribute is used for
         expression types and the empty tuple for leaf nodes.  Returning
         None is equivalent to returning (None,None).  If the callback is
         not defined, the default behavior is equivalent to returning
         (None, []).

    node_result = exitNode(self, node, data):

         exitNode() is called after the node is completely processed (as
         the walker returns up the tree to the parent node).  It is
         passed the node and the results data structure (defined by
         enterNode() and possibly further modified by
         acceptChildResult()), and is expected to return the "result" for
         this node.  If not specified, the default action is to return
         the data object from enterNode().

    descend, child_result = beforeChild(self, node, child, child_idx):

         beforeChild() is called by a node for every child before
         entering the child node.  The node, child node, and child index
         (position in the args list from enterNode()) are passed as
         arguments.  beforeChild should return a tuple (descend,
         child_result).  If descend is False, the child node will not be
         entered and the value returned to child_result will be passed to
         the node's acceptChildResult callback.  Returning None is
         equivalent to (True, None).  The default behavior if not
         specified is equivalent to (True, None).

    data = acceptChildResult(self, node, data, child_result, child_idx):

         acceptChildResult() is called for each child result being
         returned to a node.  This callback is responsible for recording
         the result for later processing or passing up the tree.  It is
         passed the node, result data structure (see enterNode()), child
         result, and the child index (position in args from enterNode()).
         The data structure (possibly modified or replaced) must be
         returned.  If acceptChildResult is not specified, it does
         nothing if data is None, otherwise it calls data.append(result).

    afterChild(self, node, child, child_idx):

         afterChild() is called by a node for every child node
         immediately after processing the node is complete before control
         moves to the next child or up to the parent node.  The node,
         child node, an child index (position in args from enterNode())
         are passed, and nothing is returned.  If afterChild is not
         specified, no action takes place.

    finalizeResult(self, result):

         finalizeResult() is called once after the entire expression tree
         has been walked.  It is passed the result returned by the root
         node exitNode() callback.  If finalizeResult is not specified,
         the walker returns the result obtained from the exitNode
         callback on the root node.

    Clients interact with this class by either deriving from it and
    implementing the necessary callbacks (see above), assigning callable
    functions to an instance of this class, or passing the callback
    functions as arguments to this class' constructor.

    """

    # The list of event methods that can either be implemented by
    # derived classes or specified as callback functions to the class
    # constructor.
    #
    # This is a dict mapping the callback name to a single character
    # that we can use to classify the set of callbacks used by a
    # particular Visitor (we define special-purpose node processors for
    # certain common combinations).  For example, a 'bex' visitor is one
    # that supports beforeChild, enterNode, and exitNode, but NOT
    # afterChild or acceptChildResult.
    client_methods = {
        'enterNode': 'e',
        'exitNode': 'x',
        'beforeChild': 'b',
        'afterChild': 'a',
        'acceptChildResult': 'c',
        'initializeWalker': '',
        'finalizeResult': '',
    }

    def __init__(self, **kwds):
        # This is slightly tricky: We want derived classes to be able to
        # override the "None" defaults here, and for keyword arguments
        # to override both.  The hasattr check prevents the "None"
        # defaults from overriding attributes or methods defined on
        # derived classes.
        for field in self.client_methods:
            if field in kwds:
                setattr(self, field, kwds.pop(field))
            elif not hasattr(self, field):
                setattr(self, field, None)
        if kwds:
            raise RuntimeError("Unrecognized keyword arguments: %s" % (kwds,))

        # Handle deprecated APIs
        _fcns = (('beforeChild', 2), ('acceptChildResult', 3), ('afterChild', 2))
        for name, nargs in _fcns:
            fcn = getattr(self, name)
            if fcn is None:
                continue
            _args = inspect.getfullargspec(fcn)
            _self_arg = 1 if inspect.ismethod(fcn) else 0
            if len(_args.args) == nargs + _self_arg and _args.varargs is None:
                deprecation_warning(
                    "Note that the API for the StreamBasedExpressionVisitor "
                    "has changed to include the child index for the %s() "
                    "method.  Please update your walker callbacks." % (name,),
                    version='5.7.0',
                )

                def wrap(fcn, nargs):
                    def wrapper(*args):
                        return fcn(*args[:nargs])

                    return wrapper

                setattr(self, name, wrap(fcn, nargs))

        self.recursion_stack = None

        # Set up the custom recursive node handler function (customized
        # for the specific set of callbacks that are defined for this
        # class instance).
        recursive_node_handler = '_process_node_' + ''.join(
            sorted(
                '' if getattr(self, f[0]) is None else f[1]
                for f in self.client_methods.items()
            )
        )
        self._process_node = getattr(
            self, recursive_node_handler, self._process_node_general
        )

    def walk_expression(self, expr):
        """Walk an expression, calling registered callbacks.

        This is the standard interface for running the visitor.  It
        defaults to using an efficient recursive implementation of the
        visitor, falling back on :py:meth:`walk_expression_nonrecursive`
        if the recursion stack gets too deep.

        """
        if self.initializeWalker is not None:
            walk, root = self.initializeWalker(expr)
            if not walk:
                return root
            elif root is None:
                root = expr
        else:
            root = expr

        try:
            result = self._process_node(root, RECURSION_LIMIT)
            _nonrecursive = None
        except RevertToNonrecursive:
            ptr = (None,) + self.recursion_stack.pop()
            while self.recursion_stack:
                ptr = (ptr,) + self.recursion_stack.pop()
            self.recursion_stack = None
            _nonrecursive = self._nonrecursive_walker_loop, ptr
        except RecursionError:
            logger.warning(
                'Unexpected RecursionError walking an expression tree.',
                extra={'id': 'W1003'},
            )
            _nonrecursive = self.walk_expression_nonrecursive, expr

        if _nonrecursive is not None:
            return _nonrecursive[0](_nonrecursive[1])

        if self.finalizeResult is not None:
            return self.finalizeResult(result)
        else:
            return result

    def _compute_actual_recursion_limit(self):
        recursion_limit = (
            sys.getrecursionlimit() - get_stack_depth() - 2 * RECURSION_LIMIT
        )
        if recursion_limit <= RECURSION_LIMIT:
            self.recursion_stack = []
            raise RevertToNonrecursive()
        return recursion_limit

    def _process_node_general(self, node, recursion_limit):
        """Recursive routine for processing nodes with general callbacks

        This is the "general" implementation of the
        StreamBasedExpressionVisitor node processor that can handle any
        combination of registered callback functions.

        """
        if not recursion_limit:
            recursion_limit = self._compute_actual_recursion_limit()
        else:
            recursion_limit -= 1

        if self.enterNode is not None:
            tmp = self.enterNode(node)
            if tmp is None:
                args = data = None
            else:
                args, data = tmp
        else:
            args = None
            data = []
        if args is None:
            if type(node) in nonpyomo_leaf_types or not node.is_expression_type():
                args = ()
            else:
                args = node.args

        # Because we do not require the args to be a context manager, we
        # will mock up the "with args" using a try-finally.
        context_manager = hasattr(args, '__enter__')
        if context_manager:
            args.__enter__()

        try:
            descend = True
            child_idx = -1
            # Note: this relies on iter(iterator) returning the
            # iterator.  This seems to hold for all common iterators
            # (list, tuple, generator, etc)
            arg_iter = iter(args)
            for child in arg_iter:
                child_idx += 1
                if self.beforeChild is not None:
                    tmp = self.beforeChild(node, child, child_idx)
                    if tmp is None:
                        descend = True
                    else:
                        descend, child_result = tmp

                if descend:
                    child_result = self._process_node(child, recursion_limit)

                if self.acceptChildResult is not None:
                    data = self.acceptChildResult(node, data, child_result, child_idx)
                elif data is not None:
                    data.append(child_result)

                if self.afterChild is not None:
                    self.afterChild(node, child, child_idx)
        except RevertToNonrecursive:
            self._recursive_frame_to_nonrecursive_stack(locals())
            context_manager = False
            raise
        finally:
            if context_manager:
                args.__exit__(None, None, None)

        # We are done with this node.  Call exitNode to compute
        # any result
        if self.exitNode is not None:
            return self.exitNode(node, data)
        else:
            return data

    def _process_node_bex(self, node, recursion_limit):
        """Recursive routine for processing nodes with only 'bex' callbacks

        This is a special-case implementation of the "general"
        StreamBasedExpressionVisitor node processor for the case that
        only beforeChild, enterNode, and exitNode are defined (see
        also the definition of the client_methods dict).

        """
        if not recursion_limit:
            recursion_limit = self._compute_actual_recursion_limit()
        else:
            recursion_limit -= 1

        tmp = self.enterNode(node)
        if tmp is None:
            args = data = None
        else:
            args, data = tmp
        if args is None:
            if type(node) in nonpyomo_leaf_types or not node.is_expression_type():
                args = ()
            else:
                args = node.args

        # Because we do not require the args to be a context manager, we
        # will mock up the "with args" using a try-finally.
        context_manager = hasattr(args, '__enter__')
        if context_manager:
            args.__enter__()

        try:
            child_idx = -1
            # Note: this relies on iter(iterator) returning the
            # iterator.  This seems to hold for all common iterators
            # (list, tuple, generator, etc)
            arg_iter = iter(args)
            for child in arg_iter:
                child_idx += 1
                tmp = self.beforeChild(node, child, child_idx)
                if tmp is None:
                    descend = True
                else:
                    descend, child_result = tmp

                if descend:
                    data.append(self._process_node(child, recursion_limit))
                else:
                    data.append(child_result)
        except RevertToNonrecursive:
            self._recursive_frame_to_nonrecursive_stack(locals())
            context_manager = False
            raise
        finally:
            if context_manager:
                args.__exit__(None, None, None)

        # We are done with this node.  Call exitNode to compute
        # any result
        return self.exitNode(node, data)

    def _process_node_bx(self, node, recursion_limit):
        """Recursive routine for processing nodes with only 'bx' callbacks

        This is a special-case implementation of the "general"
        StreamBasedExpressionVisitor node processor for the case that
        only beforeChild and exitNode are defined (see also the
        definition of the client_methods dict).

        """
        if not recursion_limit:
            recursion_limit = self._compute_actual_recursion_limit()
        else:
            recursion_limit -= 1

        if type(node) in nonpyomo_leaf_types or not node.is_expression_type():
            args = ()
        else:
            args = node.args
        data = []

        try:
            child_idx = -1
            # Note: this relies on iter(iterator) returning the
            # iterator.  This seems to hold for all common iterators
            # (list, tuple, generator, etc)
            arg_iter = iter(args)
            for child in arg_iter:
                child_idx += 1
                tmp = self.beforeChild(node, child, child_idx)
                if tmp is None:
                    descend = True
                else:
                    descend, child_result = tmp
                if descend:
                    data.append(self._process_node(child, recursion_limit))
                else:
                    data.append(child_result)
        except RevertToNonrecursive:
            self._recursive_frame_to_nonrecursive_stack(locals())
            raise
        finally:
            pass

        # We are done with this node.  Call exitNode to compute
        # any result
        return self.exitNode(node, data)

    def _recursive_frame_to_nonrecursive_stack(self, local):
        child_idx = local['child_idx']
        _arg_list = [None] * child_idx
        _arg_list.append(local['child'])
        _arg_list.extend(local['arg_iter'])
        if not self.recursion_stack:
            # For the deepest stack frame, the recursion limit hit
            # as we started to enter the child.  As we haven't
            # started processing it yet, we need to decrement
            # child_idx so that it is revisited
            child_idx -= 1
        self.recursion_stack.append(
            (local['node'], _arg_list, len(_arg_list) - 1, local['data'], child_idx)
        )

    def walk_expression_nonrecursive(self, expr):
        """Nonrecursively walk an expression, calling registered callbacks.

        This routine is safer than the recursive walkers for deep (or
        unbalanced) trees.  It is, however, slightly slower than the
        recursive implementations.

        """
        #
        # This walker uses a linked list to store the stack (instead of
        # an array).  The nodes of the linked list are 6-member tuples:
        #
        #    ( pointer to parent,
        #      expression node,
        #      tuple/list of child nodes (arguments),
        #      number of child nodes (arguments),
        #      data object to aggregate results from child nodes,
        #      current child node index )
        #
        # The walker only needs a single pointer to the end of the list
        # (ptr).  The beginning of the list is indicated by a None
        # parent pointer.
        #
        if self.initializeWalker is not None:
            walk, result = self.initializeWalker(expr)
            if not walk:
                return result
            elif result is not None:
                expr = result
        if self.enterNode is not None:
            tmp = self.enterNode(expr)
            if tmp is None:
                args = data = None
            else:
                args, data = tmp
        else:
            args = None
            data = []
        if args is None:
            if type(expr) in nonpyomo_leaf_types or not expr.is_expression_type():
                args = ()
            else:
                args = expr.args
        if hasattr(args, '__enter__'):
            args.__enter__()
        node = expr
        # Note that because we increment child_idx just before fetching
        # the child node, it must be initialized to -1, and ptr[3] must
        # always be *one less than* the number of arguments
        return self._nonrecursive_walker_loop(
            (None, node, args, len(args) - 1, data, -1)
        )

    def _nonrecursive_walker_loop(self, ptr):
        _, node, args, _, data, child_idx = ptr
        try:
            while 1:
                if child_idx < ptr[3]:
                    # Increment the child index pointer here for
                    # consistency.  Note that this means that for the bulk
                    # of the time, 'child_idx' will not match the value of
                    # ptr[5].  This provides a modest performance
                    # improvement, as we only have to recreate the ptr tuple
                    # just before we descend further into the tree (i.e., we
                    # avoid recreating the tuples for the special case where
                    # beforeChild indicates that we should not descend
                    # further).
                    child_idx += 1
                    # This node still has children to process
                    child = ptr[2][child_idx]

                    # Notify this node that we are about to descend into a
                    # child.
                    if self.beforeChild is not None:
                        tmp = self.beforeChild(node, child, child_idx)
                        if tmp is None:
                            descend = True
                            child_result = None
                        else:
                            descend, child_result = tmp
                        if not descend:
                            # We are aborting processing of this child node.
                            # Tell this node to accept the child result and
                            # we will move along
                            if self.acceptChildResult is not None:
                                data = self.acceptChildResult(
                                    node, data, child_result, child_idx
                                )
                            elif data is not None:
                                data.append(child_result)
                            # And let the node know that we are done with a
                            # child node
                            if self.afterChild is not None:
                                self.afterChild(node, child, child_idx)
                            # Jump to the top to continue processing the
                            # next child node
                            continue

                    # Update the child argument counter in the stack.
                    # Because we are using tuples, we need to recreate the
                    # "ptr" object (linked list node)
                    ptr = ptr[:4] + (data, child_idx)

                    # We are now going to actually enter this node.  The
                    # node will tell us the list of its child nodes that we
                    # need to process
                    if self.enterNode is not None:
                        tmp = self.enterNode(child)
                        if tmp is None:
                            args = data = None
                        else:
                            args, data = tmp
                    else:
                        args = None
                        data = []
                    if args is None:
                        if (
                            type(child) in nonpyomo_leaf_types
                            or not child.is_expression_type()
                        ):
                            # Leaves (either non-pyomo types or
                            # non-Expressions) have no child arguments, so
                            # are just put on the stack
                            args = ()
                        else:
                            args = child.args
                    if hasattr(args, '__enter__'):
                        args.__enter__()
                    node = child
                    child_idx = -1
                    ptr = (ptr, node, args, len(args) - 1, data, child_idx)

                else:  # child_idx == ptr[3]:
                    # We are done with this node.  Call exitNode to compute
                    # any result
                    if hasattr(ptr[2], '__exit__'):
                        ptr[2].__exit__(None, None, None)
                    if self.exitNode is not None:
                        node_result = self.exitNode(node, data)
                    else:
                        node_result = data

                    # Pop the node off the linked list
                    ptr = ptr[0]
                    # If we have returned to the beginning, return the final
                    # answer
                    if ptr is None:
                        if self.finalizeResult is not None:
                            return self.finalizeResult(node_result)
                        else:
                            return node_result
                    # Not done yet, update node to point to the new active
                    # node
                    node, child = ptr[1], node
                    data = ptr[4]
                    child_idx = ptr[5]

                    # We need to alert the node to accept the child's result:
                    if self.acceptChildResult is not None:
                        data = self.acceptChildResult(
                            node, data, node_result, child_idx
                        )
                    elif data is not None:
                        data.append(node_result)

                    # And let the node know that we are done with a child node
                    if self.afterChild is not None:
                        self.afterChild(node, child, child_idx)

        finally:
            while ptr is not None:
                if hasattr(ptr[2], '__exit__'):
                    ptr[2].__exit__(None, None, None)
                ptr = ptr[0]


@deprecated(
    "The SimpleExpressionVisitor is deprecated.  "
    "Please use the StreamBasedExpressionVisitor instead.",
    version='6.9.0',
)
class SimpleExpressionVisitor:
    """
    Note:
        This class is a customization of the PyUtilib :class:`SimpleVisitor
        <pyutilib.misc.visitor.SimpleVisitor>` class that is tailored
        to efficiently walk Pyomo expression trees.  However, this class
        is not a subclass of the PyUtilib :class:`SimpleVisitor
        <pyutilib.misc.visitor.SimpleVisitor>` class because all key methods
        are reimplemented.
    """

    def visit(self, node):  # pragma: no cover
        """
        Visit a node in an expression tree and perform some operation on
        it.

        This method should be over-written by a user
        that is creating a sub-class.

        Args:
            node: a node in an expression tree

        Returns:
            nothing
        """
        pass

    def finalize(self):  # pragma: no cover
        """
        Return the "final value" of the search.

        The default implementation returns :const:`None`, because
        the traditional visitor pattern does not return a value.

        Returns:
            The final value after the search.  Default is :const:`None`.
        """
        pass

    def xbfs(self, node):
        """
        Breadth-first search of an expression tree,
        except that leaf nodes are immediately visited.

        Note:
            This method has the same functionality as the
            PyUtilib :class:`SimpleVisitor.xbfs <pyutilib.misc.visitor.SimpleVisitor.xbfs>`
            method.  The difference is that this method
            is tailored to efficiently walk Pyomo expression trees.

        Args:
            node: The root node of the expression tree that is searched.

        Returns:
            The return value is determined by the :func:`finalize` function,
            which may be defined by the user.  Defaults to :const:`None`.
        """
        if (
            node.__class__ in nonpyomo_leaf_types
            or not node.is_expression_type()
            or node.nargs() == 0
        ):
            self.visit(node)
            return self.finalize()

        dq = deque([node])
        while dq:
            current = dq.popleft()
            self.visit(current)
            # for c in self.children(current):
            for c in current.args:
                # if self.is_leaf(c):
                if (
                    c.__class__ in nonpyomo_leaf_types
                    or not c.is_expression_type()
                    or c.nargs() == 0
                ):
                    self.visit(c)
                else:
                    dq.append(c)
        return self.finalize()

    def xbfs_yield_leaves(self, node):
        """
        Breadth-first search of an expression tree, except that
        leaf nodes are immediately visited.

        Note:
            This method has the same functionality as the
            PyUtilib :class:`SimpleVisitor.xbfs_yield_leaves <pyutilib.misc.visitor.SimpleVisitor.xbfs_yield_leaves>`
            method.  The difference is that this method
            is tailored to efficiently walk Pyomo expression trees.

        Args:
            node: The root node of the expression tree
                that is searched.

        Returns:
            The return value is determined by the :func:`finalize` function,
            which may be defined by the user.  Defaults to :const:`None`.
        """
        #
        # If we start with a leaf, then yield it and stop iteration
        #
        if (
            node.__class__ in nonpyomo_leaf_types
            or not node.is_expression_type()
            or node.nargs() == 0
        ):
            ans = self.visit(node)
            if not ans is None:
                yield ans
            return
        #
        # Iterate through the tree.
        #
        dq = deque([node])
        while dq:
            current = dq.popleft()
            # self.visit(current)
            # for c in self.children(current):
            for c in current.args:
                # if self.is_leaf(c):
                if (
                    c.__class__ in nonpyomo_leaf_types
                    or not c.is_expression_type()
                    or c.nargs() == 0
                ):
                    ans = self.visit(c)
                    if not ans is None:
                        yield ans
                else:
                    dq.append(c)


class ExpressionValueVisitor:
    """
    Note:
        This class is a customization of the PyUtilib :class:`ValueVisitor
        <pyutilib.misc.visitor.ValueVisitor>` class that is tailored
        to efficiently walk Pyomo expression trees.  However, this class
        is not a subclass of the PyUtilib :class:`ValueVisitor
        <pyutilib.misc.visitor.ValueVisitor>` class because all key methods
        are reimplemented.
    """

    def visit(self, node, values):  # pragma: no cover
        """
        Visit a node in a tree and compute its value using
        the values of its children.

        This method should be over-written by a user
        that is creating a sub-class.

        Args:
            node: a node in a tree
            values: a list of values of this node's children

        Returns:
            The *value* for this node, which is computed using :attr:`values`
        """
        pass

    def visiting_potential_leaf(self, node):  # pragma: no cover
        """
        Visit a node and return its value if it is a leaf.

        Note:
            This method needs to be over-written for a specific
            visitor application.

        Args:
            node: a node in a tree

        Returns:
            A tuple: ``(flag, value)``.   If ``flag`` is False,
            then the node is not a leaf and ``value`` is :const:`None`.
            Otherwise, ``value`` is the computed value for this node.
        """
        raise RuntimeError("The visiting_potential_leaf method needs to be defined.")

    def finalize(self, ans):  # pragma: no cover
        """
        This method defines the return value for the search methods
        in this class.

        The default implementation returns the value of the
        initial node (aka the root node), because
        this visitor pattern computes and returns value for each
        node to enable the computation of this value.

        Args:
            ans: The final value computed by the search method.

        Returns:
            The final value after the search. Defaults to simply
            returning :attr:`ans`.
        """
        return ans

    def dfs_postorder_stack(self, node):
        """
        Perform a depth-first search in postorder using a stack
        implementation.

        Note:
            This method has the same functionality as the
            PyUtilib :class:`ValueVisitor.dfs_postorder_stack <pyutilib.misc.visitor.ValueVisitor.dfs_postorder_stack>`
            method.  The difference is that this method
            is tailored to efficiently walk Pyomo expression trees.

        Args:
            node: The root node of the expression tree
                that is searched.

        Returns:
            The return value is determined by the :func:`finalize` function,
            which may be defined by the user.
        """
        flag, value = self.visiting_potential_leaf(node)
        if flag:
            return self.finalize(value)
        # _stack = [ (node, self.children(node), 0, len(self.children(node)), [])]
        _stack = [(node, node.args, 0, node.nargs(), [])]
        #
        # Iterate until the stack is empty
        #
        # Note: 1 is faster than True for Python 2.x
        #
        while 1:
            #
            # Get the top of the stack
            #   _obj        Current expression object
            #   _argList    The arguments for this expression object
            #   _idx        The current argument being considered
            #   _len        The number of arguments
            #   _result     The return values
            #
            _obj, _argList, _idx, _len, _result = _stack.pop()
            #
            # Iterate through the arguments
            #
            while _idx < _len:
                _sub = _argList[_idx]
                _idx += 1
                flag, value = self.visiting_potential_leaf(_sub)
                if flag:
                    _result.append(value)
                else:
                    #
                    # Push an expression onto the stack
                    #
                    _stack.append((_obj, _argList, _idx, _len, _result))
                    _obj = _sub
                    # _argList                = self.children(_sub)
                    _argList = _sub.args
                    _idx = 0
                    _len = _sub.nargs()
                    _result = []
            #
            # Process the current node
            #
            ans = self.visit(_obj, _result)
            if _stack:
                #
                # "return" the recursion by putting the return value on
                # the end of the results stack
                #
                _stack[-1][-1].append(ans)
            else:
                return self.finalize(ans)


def replace_expressions(
    expr,
    substitution_map,
    descend_into_named_expressions=True,
    remove_named_expressions=True,
):
    """

    Parameters
    ----------
    expr : Pyomo expression
       The source expression
    substitution_map : dict
       A dictionary mapping object ids in the source to the replacement objects.
    descend_into_named_expressions : bool
       True if replacement should go into named expression objects, False to halt at
       a named expression
    remove_named_expressions : bool
       True if the named expressions should be replaced with a standard expression,
       and False if the named expression should be left in place

    Returns
    -------
       Pyomo expression : returns the new expression object
    """
    return ExpressionReplacementVisitor(
        substitute=substitution_map,
        descend_into_named_expressions=descend_into_named_expressions,
        remove_named_expressions=remove_named_expressions,
    ).walk_expression(expr)


class ExpressionReplacementVisitor(StreamBasedExpressionVisitor):
    def __init__(
        self,
        substitute=None,
        descend_into_named_expressions=True,
        remove_named_expressions=True,
    ):
        if substitute is None:
            substitute = {}
        # Note: preserving the attribute names from the previous
        # implementation of the expression walker.
        self.substitute = substitute
        self.enter_named_expr = descend_into_named_expressions
        self.rm_named_expr = remove_named_expressions

        kwds = {}
        if hasattr(self, 'visiting_potential_leaf'):
            deprecation_warning(
                "ExpressionReplacementVisitor: this walker has been ported "
                "to derive from StreamBasedExpressionVisitor.  "
                "visiting_potential_leaf() has been replaced by beforeChild()"
                "(note to implementers: the sense of the bool return value "
                "has been inverted).",
                version='6.2',
            )

            def beforeChild(node, child, child_idx):
                is_leaf, ans = self.visiting_potential_leaf(child)
                return not is_leaf, ans

            kwds['beforeChild'] = beforeChild

        if hasattr(self, 'visit'):
            raise DeveloperError(
                "ExpressionReplacementVisitor: this walker has been ported "
                "to derive from StreamBasedExpressionVisitor.  "
                "overriding visit() has no effect (and is likely to generate "
                "invalid expression trees)"
            )
        super().__init__(**kwds)

    def initializeWalker(self, expr):
        walk, result = self.beforeChild(None, expr, 0)
        if not walk:
            return False, result
        return True, expr

    def beforeChild(self, node, child, child_idx):
        if id(child) in self.substitute:
            return False, self.substitute[id(child)]
        elif type(child) in native_types:
            return False, child
        elif not child.is_expression_type():
            return False, child
        elif child.is_named_expression_type():
            if not self.enter_named_expr:
                return False, child
        return True, None

    def enterNode(self, node):
        args = list(node.args)
        # [bool:args_have_changed, list:original_args, bool:node_is_constant]
        return args, [False, args, True]

    def acceptChildResult(self, node, data, child_result, child_idx):
        if data[1][child_idx] is not child_result:
            data[1][child_idx] = child_result
            data[0] = True
        if (
            child_result.__class__ not in native_types
            and not child_result.is_constant()
        ):
            data[2] = False
        return data

    def exitNode(self, node, data):
        if node.is_named_expression_type():
            assert len(data[1]) == 1
            if self.rm_named_expr:
                return data[1][0]
            elif data[0]:
                node.set_value(data[1][0])
                return node
        elif data[0]:
            if data[2]:
                return node._apply_operation(data[1])
            else:
                return node.create_node_with_local_data(data[1])
        return node

    @deprecated(
        "ExpressionReplacementVisitor: this walker has been ported "
        "to derive from StreamBasedExpressionVisitor.  "
        "dfs_postorder_stack() has been replaced with walk_expression()",
        version='6.2',
    )
    def dfs_postorder_stack(self, expr):
        return self.walk_expression(expr)


def evaluate_fixed_subexpressions(
    expr, descend_into_named_expressions=True, remove_named_expressions=True
):
    return EvaluateFixedSubexpressionVisitor(
        descend_into_named_expressions=descend_into_named_expressions,
        remove_named_expressions=remove_named_expressions,
    ).walk_expression(expr)


class EvaluateFixedSubexpressionVisitor(ExpressionReplacementVisitor):
    def __init__(
        self, descend_into_named_expressions=False, remove_named_expressions=False
    ):
        super().__init__(
            descend_into_named_expressions=descend_into_named_expressions,
            remove_named_expressions=remove_named_expressions,
        )

    def beforeChild(self, node, child, child_idx):
        if type(child) in native_types:
            return False, child
        elif not child.is_expression_type():
            if child.is_fixed():
                return False, child()
            else:
                return False, child
        elif child.is_named_expression_type():
            if not self.enter_named_expr:
                return False, child
        return True, None


# -------------------------------------------------------
#
# Functions used to process expression trees
#
# -------------------------------------------------------

# =====================================================
#  clone_expression
# =====================================================


def clone_expression(expr, substitute=None):
    """A function that is used to clone an expression.

    Cloning is equivalent to calling ``copy.deepcopy`` with no Block
    scope.  That is, the expression tree is duplicated, but no Pyomo
    components (leaf nodes *or* named Expressions) are duplicated.

    Args:
        expr: The expression that will be cloned.
        substitute (dict): A dictionary mapping object ids to
            objects. This dictionary has the same semantics as
            the memo object used with ``copy.deepcopy``. Defaults
            to None, which indicates that no user-defined
            dictionary is used.

    Returns:
        The cloned expression.

    """
    common.clone_counter._count += 1
    memo = {'__block_scope__': {id(None): False}}
    if substitute:
        expr = replace_expressions(expr, substitute)
    return deepcopy(expr, memo)


# =====================================================
#  sizeof_expression
# =====================================================


def sizeof_expression(expr):
    """
    Return the number of nodes in the expression tree.

    Args:
        expr: The root node of an expression tree.

    Returns:
        A non-negative integer that is the number of
        interior and leaf nodes in the expression tree.
    """

    def enter(node):
        return None, 1

    def accept(node, data, child_result, child_idx):
        return data + child_result

    return StreamBasedExpressionVisitor(
        enterNode=enter, acceptChildResult=accept
    ).walk_expression(expr)


# =====================================================
#  evaluate_expression
# =====================================================


class _EvaluationVisitor(ExpressionValueVisitor):
    def __init__(self, exception):
        self.exception = exception

    def visit(self, node, values):
        """Visit nodes that have been expanded"""
        return node._apply_operation(values)

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node.__class__ in nonpyomo_leaf_types:
            return True, node

        if node.is_expression_type():
            return False, None

        if node.is_numeric_type():
            return True, value(node, exception=self.exception)
        elif node.is_logical_type():
            return True, value(node, exception=self.exception)
        else:
            return True, node


class FixedExpressionError(Exception):
    def __init__(self, *args, **kwds):
        super(FixedExpressionError, self).__init__(*args, **kwds)


class NonConstantExpressionError(Exception):
    def __init__(self, *args, **kwds):
        super(NonConstantExpressionError, self).__init__(*args, **kwds)


class _EvaluateConstantExpressionVisitor(ExpressionValueVisitor):
    def visit(self, node, values):
        """Visit nodes that have been expanded"""
        return node._apply_operation(values)

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node.__class__ in nonpyomo_leaf_types:
            return True, node

        if node.is_expression_type():
            return False, None

        if node.is_numeric_type():
            # Get the object value.  This will also cause templates to
            # raise TemplateExpressionErrors
            try:
                val = value(node)
            except TemplateExpressionError:
                raise
            except:
                # Uninitialized Var/Param objects should be given the
                # opportunity to map the error to a NonConstant / Fixed
                # expression error
                if not node.is_fixed():
                    raise NonConstantExpressionError(
                        f"{node} ({type(node).__name__}) is not fixed"
                    )
                if not node.is_constant():
                    raise FixedExpressionError(
                        f"{node} ({type(node).__name__}) is not constant"
                    )
                raise

            if not node.is_fixed():
                raise NonConstantExpressionError()
            if not node.is_constant():
                raise FixedExpressionError()
            return True, val

        return True, node


def evaluate_expression(exp, exception=True, constant=False):
    """Evaluate the value of the expression.

    Args:
        expr: The root node of an expression tree.
        exception (bool): A flag that indicates whether
            exceptions are raised.  If this flag is
            :const:`False`, then an exception that
            occurs while evaluating the expression
            is caught and the return value is :const:`None`.
            Default is :const:`True`.
        constant (bool): If True, constant expressions are
            evaluated and returned but nonconstant expressions
            raise either FixedExpressionError or
            NonconstantExpressionError (default=False).

    Returns:
        A floating point value if the expression evaluates
        normally, or :const:`None` if an exception occurs
        and is caught.

    """
    clear_active = False
    if constant:
        visitor = _EvaluateConstantExpressionVisitor()
    else:
        if evaluate_expression.visitor_active:
            visitor = _EvaluationVisitor(exception=exception)
        else:
            visitor = evaluate_expression.visitor_cache
            visitor.exception = exception
            evaluate_expression.visitor_active = True
            clear_active = True

    try:
        ans = visitor.dfs_postorder_stack(exp)
    except (
        TemplateExpressionError,
        ValueError,
        TypeError,
        NonConstantExpressionError,
        FixedExpressionError,
    ):
        # Errors that we want to be able to suppress:
        #
        #   TemplateExpressionError: raised when generating expression
        #      templates
        #   FixedExpressionError, NonConstantExpressionError: raised
        #      when processing expressions that are expected to be fixed
        #      (e.g., indices)
        #   ValueError: "standard" expression value errors
        #   TypeError: This can be raised in Python3 when evaluating a
        #      operation returns a complex number (e.g., sqrt(-1))
        if exception:
            raise
        return None
    finally:
        if clear_active:
            evaluate_expression.visitor_active = False
    if ans.__class__ not in native_types and ans.is_numeric_type() is True:
        return value(ans)
    return ans


evaluate_expression.visitor_cache = _EvaluationVisitor(True)
evaluate_expression.visitor_active = False

# =====================================================
#  identify_components
# =====================================================


class _ComponentVisitor(StreamBasedExpressionVisitor):
    def __init__(self, types):
        super().__init__()
        if types.__class__ is not set:
            types = set(types)
        self._types = types

    def initializeWalker(self, expr):
        self._objs = []
        self._seen = set()
        return True, None

    def finalizeResult(self, result):
        return self._objs

    def exitNode(self, node, data):
        if node.__class__ in self._types and id(node) not in self._seen:
            self._seen.add(id(node))
            self._objs.append(node)


def identify_components(expr, component_types):
    """
    A generator that yields a sequence of nodes
    in an expression tree that belong to a specified set.

    Args:
        expr: The root node of an expression tree.
        component_types (set or list): A set of class
            types that will be matched during the search.

    Yields:
        Each node that is found.
    """
    #
    # OPTIONS:
    # component_types - set (or list) if class types to find
    # in the expression.
    #
    visitor = _ComponentVisitor(component_types)
    yield from visitor.walk_expression(expr)


# =====================================================
#  identify_variables
# =====================================================


class IdentifyVariableVisitor(StreamBasedExpressionVisitor):
    def __init__(self, include_fixed=False, named_expression_cache=None):
        """Visitor that collects all unique variables participating in an
        expression

        Args:
            include_fixed (bool): Whether to include fixed variables
            named_expression_cache (optional, dict): Dict mapping ids of named
                expressions to a tuple of the list of all variables and the
                set of all variable ids contained in the named expression.

        """
        super().__init__()
        self._include_fixed = include_fixed
        self._cache = named_expression_cache
        # Stack of named expressions. This holds the tuple
        #     (eid, _seen, _exprs)
        # where eid is the id() of the subexpression we are currently
        # processing, and _seen and _exprs are from the parent context.
        self._expr_stack = []
        # The following attributes will be added by initializeWalker:
        # self._seen: dict(eid: obj)
        # self._exprs: list of (e, e.expr) for any (nested) named expressions

    def initializeWalker(self, expr):
        assert not self._expr_stack
        self._seen = {}
        self._exprs = None
        if not self.beforeChild(None, expr, 0)[0]:
            return False, self.finalizeResult(None)
        return True, expr

    def beforeChild(self, parent, child, index):
        if child.__class__ in native_types:
            return False, None
        elif child.is_expression_type():
            if child.is_named_expression_type():
                return self._process_named_expr(child)
            else:
                return True, None
        elif child.is_variable_type() and (self._include_fixed or not child.fixed):
            if id(child) not in self._seen:
                self._seen[id(child)] = child
        return False, None

    def exitNode(self, node, data):
        if node.is_named_expression_type() and self._cache is not None:
            # If we are returning from a named expression, we must make
            # sure that we properly restore the "outer" context and then
            # merge the objects from the named expression we just exited
            # into the list for the parent expression context.
            _seen = self._seen
            _exprs = self._exprs
            eid, self._seen, self._exprs = self._expr_stack.pop()
            assert eid == id(node)
            self._merge_obj_lists(_seen, _exprs)

    def finalizeResult(self, result):
        assert not self._expr_stack
        return self._seen.values()

    def _merge_obj_lists(self, _seen, _exprs):
        self._seen.update(_seen)
        if self._exprs is not None:
            self._exprs.update(_exprs)

    def _process_named_expr(self, child):
        if self._cache is None:
            return True, None
        eid = id(child)
        if eid in self._cache:
            _seen, _exprs = self._cache[eid]
            if all(c.expr is e for c, e in _exprs.values()):
                # We have already encountered this named expression. We just add
                # the cached objects to our list and don't descend.
                #
                # Note that a cache hit requires not only that we have seen
                # this expression before, but also that none of the named
                # expressions have changed.  If they have, then the cache
                # miss will fall over to the else clause below and descend
                # into the expression, (implicitly) rebuilding the cache.
                self._merge_obj_lists(_seen, _exprs)
                return False, None
        # If we are descending into a new named expression or a cached
        # named expression where the cache is now invalid.  Initialize a
        # cache to store the expression's local objects.
        self._expr_stack.append((eid, self._seen, self._exprs))
        self._seen = {}
        self._exprs = {eid: (child, child.expr)}
        self._cache[eid] = (self._seen, self._exprs)
        return True, None


def identify_variables(expr, include_fixed=True, named_expression_cache=None):
    """
    A generator that yields a sequence of variables
    in an expression tree.

    Args:
        expr: The root node of an expression tree.
        include_fixed (bool): If :const:`True`, then
            this generator will yield variables whose
            value is fixed.  Defaults to :const:`True`.

    Yields:
        Each variable that is found.
    """
    v = identify_variables.visitor
    save = v._include_fixed, v._cache
    try:
        v._include_fixed = include_fixed
        v._cache = named_expression_cache
        yield from v.walk_expression(expr)
    finally:
        v._include_fixed, v._cache = save


identify_variables.visitor = IdentifyVariableVisitor()


# =====================================================
#  identify_mutable_parameters
# =====================================================


class IdentifyMutableParamVisitor(IdentifyVariableVisitor):
    def __init__(self):
        # Hide the IdentifyVariableVisitor API (not relevant here)
        super().__init__()

    def beforeChild(self, parent, child, index):
        if child.__class__ in native_types:
            return False, None
        elif child.is_expression_type():
            if child.is_named_expression_type():
                return self._process_named_expr(child)
            else:
                return True, None
        if (
            not child.is_variable_type()
            and child.is_fixed()
            and not child.is_constant()
        ):
            if id(child) not in self._seen:
                self._seen[id(child)] = child
        return False, None


def identify_mutable_parameters(expr):
    """
    A generator that yields a sequence of mutable
    parameters in an expression tree.

    Args:
        expr: The root node of an expression tree.

    Yields:
        Each mutable parameter that is found.
    """
    yield from identify_mutable_parameters.visitor.walk_expression(expr)


identify_mutable_parameters.visitor = IdentifyMutableParamVisitor()

# =====================================================
#  polynomial_degree
# =====================================================


class _PolynomialDegreeVisitor(ExpressionValueVisitor):
    def visit(self, node, values):
        """Visit nodes that have been expanded"""
        return node._compute_polynomial_degree(values)

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node.__class__ in nonpyomo_leaf_types:
            return True, 0

        if node.is_expression_type():
            return False, None

        if node.is_numeric_type():
            return True, 0 if node.is_fixed() else 1
        else:
            return True, node


def polynomial_degree(node):
    """
    Return the polynomial degree of the expression.

    Args:
        node: The root node of an expression tree.

    Returns:
        A non-negative integer that is the polynomial
        degree if the expression is polynomial, or :const:`None` otherwise.
    """
    visitor = _PolynomialDegreeVisitor()
    return visitor.dfs_postorder_stack(node)


# =====================================================
#  _expression_is_fixed
# =====================================================


class _IsFixedVisitor(ExpressionValueVisitor):
    """
    NOTE: This doesn't check if combiner logic is
    all or any and short-circuit the test.  It's
    not clear that that is an important optimization.
    """

    def visit(self, node, values):
        """Visit nodes that have been expanded"""
        return node._is_fixed(values)

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node.__class__ in nonpyomo_leaf_types:
            return True, True

        elif node.is_expression_type():
            return False, None

        elif node.is_numeric_type():
            return True, node.is_fixed()

        return True, node


def _expression_is_fixed(node):
    """Return bool indicating if this expression is fixed (non-variable)

    Args:
        node: The root node of an expression tree.

    Returns: bool

    """
    visitor = _IsFixedVisitor()
    return visitor.dfs_postorder_stack(node)


# =====================================================
#  expression_to_string
# =====================================================

LEFT_TO_RIGHT = common.OperatorAssociativity.LEFT_TO_RIGHT
RIGHT_TO_LEFT = common.OperatorAssociativity.RIGHT_TO_LEFT


class _ToStringVisitor(ExpressionValueVisitor):
    _expression_handlers = None
    _leaf_node_types = set()

    def __init__(self, verbose, smap):
        super(_ToStringVisitor, self).__init__()
        self.verbose = verbose
        self.smap = smap

    def visit(self, node, values):
        """Visit nodes that have been expanded"""
        node_prec = node.PRECEDENCE
        if node_prec is not None and not self.verbose:
            for i, (val, arg) in enumerate(zip(values, node.args)):
                arg_prec = getattr(arg, 'PRECEDENCE', None)
                if arg_prec is None:
                    # This embedded constant (4) is evil, but to actually
                    # import the NegationExpression.PRECEDENCE from
                    # numeric_expr would create a circular dependency.
                    #
                    # FIXME: rework the dependencies between
                    # numeric_expr and visitor
                    if val[0] == '-' and node_prec < 4:
                        values[i] = f"({val})"
                else:
                    if node_prec < arg_prec:
                        parens = True
                    elif node_prec == arg_prec:
                        if i == 0:
                            parens = node.ASSOCIATIVITY != LEFT_TO_RIGHT
                        elif i == node.nargs() - 1:
                            parens = node.ASSOCIATIVITY != RIGHT_TO_LEFT
                        else:
                            parens = True
                    else:
                        parens = False
                    if parens:
                        values[i] = f"({val})"

        if self._expression_handlers and node.__class__ in self._expression_handlers:
            return self._expression_handlers[node.__class__](self, node, values)

        return node._to_string(values, self.verbose, self.smap)

    def visiting_potential_leaf(self, node):
        """
        Visiting a potential leaf.

        Return True if the node is not expanded.
        """
        if node is None:
            return True, 'Undefined'

        if node.__class__ in native_numeric_types:
            return True, str(node)

        if node.__class__ in nonpyomo_leaf_types:
            return True, repr(node)

        if node.is_expression_type() and node.__class__ not in self._leaf_node_types:
            return False, None

        if hasattr(node, 'to_string'):
            return True, node.to_string(verbose=self.verbose, smap=self.smap)
        elif self.smap is not None:
            return True, self.smap.getSymbol(node)
        else:
            return True, str(node)


def expression_to_string(
    expr, verbose=None, labeler=None, smap=None, compute_values=False
):
    """Return a string representation of an expression.

    Parameters
    ----------
    expr: ExpressionBase
        The root node of an expression tree.

    verbose: bool
        If :const:`True`, then the output is a nested functional form.
        Otherwise, the output is an algebraic expression.  Default is
        retrieved from :py:attr:`common.TO_STRING_VERBOSE`

    labeler: Callable
        If specified, this labeler is used to generate the string
        representation for leaves (Var / Param objects) in the
        expression.

    smap:  SymbolMap
        If specified, this :class:`SymbolMap
        <pyomo.core.expr.symbol_map.SymbolMap>` is used to cache labels.

    compute_values: bool
        If :const:`True`, then parameters and fixed variables are
        evaluated before the expression string is generated.  Default is
        :const:`False`.

    Returns:
        A string representation for the expression.

    """
    verbose = common.TO_STRING_VERBOSE if verbose is None else verbose
    #
    # Setup the symbol map
    #
    if labeler is not None:
        if smap is None:
            smap = SymbolMap()
        smap.default_labeler = labeler
    #
    # TODO: should we deprecate the compute_values option?
    #
    if compute_values:
        expr = evaluate_fixed_subexpressions(expr)
    #
    # Create and execute the visitor pattern
    #
    visitor = _ToStringVisitor(verbose, smap)
    return visitor.dfs_postorder_stack(expr)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.deprecation import deprecated
import pyomo.core.expr.calculus.derivatives as diff_core
from pyomo.core.expr.calculus.diff_with_sympy import differentiate_available
from pyomo.common.errors import NondifferentiableError


@deprecated(
    msg=(
        'The differentiate function in pyomo.core.base.symbolic has been deprecated. Please use the '
        + 'differentiate function in pyomo.core.expr.'
    ),
    version='5.6.7',
    remove_in='5.7',
)
def differentiate(expr, wrt=None, wrt_list=None):
    """Return derivative of expression.

    This function returns an expression or list of expression objects
    corresponding to the derivative of the passed expression 'expr' with
    respect to a variable 'wrt' or list of variables 'wrt_list'

    Args:
        expr (Expression): Pyomo expression
        wrt (Var): Pyomo variable
        wrt_list (list): list of Pyomo variables

    Returns:
        Expression or list of Expression objects

    """
    return diff_core.differentiate(
        expr=expr, wrt=wrt, wrt_list=wrt_list, mode=diff_core.Modes.sympy
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from __future__ import annotations
import logging
import sys
from pyomo.common.pyomo_typing import overload
from weakref import ref as weakref_ref
from typing import Union, Type

from pyomo.common.deprecation import RenamedClass
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import NOTSET
from pyomo.common.timing import ConstructionTimer

from pyomo.core.staleflag import StaleFlagManager
from pyomo.core.expr import GetItemExpression
from pyomo.core.expr.numeric_expr import NPV_MaxExpression, NPV_MinExpression
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.numvalue import (
    NumericValue,
    value,
    is_potentially_variable,
    native_numeric_types,
)
from pyomo.core.base.component import ComponentData, ModelComponentFactory
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.disable_methods import disable_methods
from pyomo.core.base.indexed_component import (
    IndexedComponent,
    UnindexedComponent_set,
    IndexedComponent_NDArrayMixin,
)
from pyomo.core.base.initializer import (
    Initializer,
    DefaultInitializer,
    BoundInitializer,
)
from pyomo.core.base.set import (
    Reals,
    Binary,
    Set,
    SetInitializer,
    real_global_set_ids,
    integer_global_set_ids,
)
from pyomo.core.base.units_container import units

logger = logging.getLogger('pyomo.core')

_inf = float('inf')
_ninf = -_inf
_nonfinite_values = {_inf, _ninf}
_known_global_real_domains = dict(
    [(_, True) for _ in real_global_set_ids]
    + [(_, False) for _ in integer_global_set_ids]
)
_VARDATA_API = (
    'domain',
    'bounds',
    'lower',
    'upper',
    'lb',
    'ub',
    'has_lb',
    'has_ub',
    'setlb',
    'setub',
    'get_units',
    'is_integer',
    'is_binary',
    'is_continuous',
    'is_fixed',
    'fix',
    'unfix',
    'free',
    'set_value',
    'value',
    'stale',
    'fixed',
    ('__call__', "access property 'value' on"),
)


class VarData(ComponentData, NumericValue):
    """This class defines the data for a single variable."""

    __slots__ = ('_value', '_lb', '_ub', '_domain', '_fixed', '_stale')
    __autoslot_mappers__ = {'_stale': StaleFlagManager.stale_mapper}

    def __init__(self, component=None):
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - VarData
        #   - ComponentData
        #   - NumericValue
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = NOTSET
        self._value = None
        #
        # The type of the lower and upper bound attributes can either be
        # atomic numeric types in Python, expressions, etc.  Basically,
        # they can be anything that passes an "not
        # is_potentially_variable" test.
        #
        self._lb = None
        self._ub = None
        self._domain = None
        self._fixed = False
        self._stale = 0  # True

    @classmethod
    def copy(cls, src):
        self = cls.__new__(cls)
        self._component = src._component
        self._value = src._value
        self._lb = src._lb
        self._ub = src._ub
        self._domain = src._domain
        self._fixed = src._fixed
        self._stale = src._stale
        self._index = src._index
        return self

    def set_value(self, val, skip_validation=False):
        """Set the current variable value.

        Set the value of this variable.  The incoming value is converted
        to a numeric value (i.e., expressions are evaluated).  If the
        variable has units, the incoming value is converted to the
        correct units before storing the value.  The final value is
        checked against both the variable domain and bounds, and an
        exception is raised if the value is not valid.  Domain and
        bounds checking can be bypassed by setting the ``skip_validation``
        argument to :const:`True`.

        """
        # Special case: setting a variable to None "clears" the variable.
        if val is None:
            self._value = None
            self._stale = 0  # True
            return
        # TODO: generate a warning/error:
        #
        # Check if this Var has units: assigning dimensionless
        # values to a variable with units should be an error
        if val.__class__ in native_numeric_types:
            pass
        elif self.parent_component()._units is not None:
            _src_magnitude = value(val)
            # Note: value() could have just registered a new numeric type
            if val.__class__ in native_numeric_types:
                val = _src_magnitude
            else:
                _src_units = units.get_units(val)
                val = units.convert_value(
                    num_value=_src_magnitude,
                    from_units=_src_units,
                    to_units=self.parent_component()._units,
                )
        else:
            val = value(val)

        if not skip_validation:
            if val not in self.domain:
                logger.warning(
                    "Setting Var '%s' to a value `%s` (%s) not in domain %s."
                    % (self.name, val, type(val).__name__, self.domain),
                    extra={'id': 'W1001'},
                )
            elif (self._lb is not None and val < value(self._lb)) or (
                self._ub is not None and val > value(self._ub)
            ):
                logger.warning(
                    "Setting Var '%s' to a numeric value `%s` "
                    "outside the bounds %s." % (self.name, val, self.bounds),
                    extra={'id': 'W1002'},
                )

        self._value = val
        self._stale = StaleFlagManager.get_flag(self._stale)

    @property
    def value(self):
        """Return (or set) the value for this variable."""
        return self._value

    @value.setter
    def value(self, val):
        self.set_value(val)

    def __call__(self, exception=NOTSET):
        """Compute the value of this variable."""
        exception = _type_check_exception_arg(self, exception)
        return self._value

    @property
    def domain(self):
        """Return (or set) the domain for this variable."""
        return self._domain

    @domain.setter
    def domain(self, domain):
        try:
            self._domain = SetInitializer(domain)(
                self.parent_block(), self.index(), self
            )
        except:
            logger.error(
                "%s is not a valid domain. Variable domains must be an "
                "instance of a Pyomo Set or convertible to a Pyomo Set." % (domain,),
                extra={'id': 'E2001'},
            )
            raise

    def has_lb(self):
        """Returns :const:`False` when the lower bound is
        :const:`None` or negative infinity"""
        return self.lb is not None

    def has_ub(self):
        """Returns :const:`False` when the upper bound is
        :const:`None` or positive infinity"""
        return self.ub is not None

    # TODO: deprecate this?  Properties are generally preferred over "set*()"
    def setlb(self, val):
        """
        Set the lower bound for this variable after validating that
        the value is fixed (or None).
        """
        self.lower = val

    # TODO: deprecate this?  Properties are generally preferred over "set*()"
    def setub(self, val):
        """
        Set the upper bound for this variable after validating that
        the value is fixed (or None).
        """
        self.upper = val

    def _resolve_bound_value(self, bound, domain_bound, unbounded, resolver):
        if bound is None:
            return domain_bound

        if bound.__class__ not in native_numeric_types:
            bound = value(bound)
            if bound.__class__ not in native_numeric_types:
                # Starting in numpy 1.25, casting 1-element ndarray to
                # float is deprecated.  We still want to support
                # that... but without enforcing a hard numpy dependence
                for cls in bound.__class__.__mro__:
                    if cls.__name__ == 'ndarray' and cls.__module__ == 'numpy':
                        if len(bound) == 1:
                            bound = bound[0]
                        break
                bound = float(bound)
        if bound in _nonfinite_values or bound != bound:
            if bound == unbounded:
                bound = None
            else:
                raise ValueError(
                    "Var '%s' created with an invalid non-finite "
                    "%s bound (%s)."
                    % (self.name, "lower" if unbounded == _ninf else "upper", bound)
                )
        if domain_bound is not None:
            if bound is None:
                return domain_bound
            else:
                return resolver(bound, domain_bound)
        return bound

    @property
    def bounds(self):
        """Returns (or set) the tuple (lower bound, upper bound).

        This returns the current (numeric) values of the lower and upper
        bounds as a tuple.  If there is no bound, returns None (and not
        +/-inf)

        """
        # Custom implementation of lb / ub to avoid unnecessary
        # expression generation and duplicate calls to domain.bounds()
        domain_lb, domain_ub = self.domain.bounds()
        # return the tighter of the domain and bounds
        return (
            self._resolve_bound_value(self._lb, domain_lb, _ninf, max),
            self._resolve_bound_value(self._ub, domain_ub, _inf, min),
        )

    @bounds.setter
    def bounds(self, val):
        self.lower, self.upper = val

    @property
    def lb(self):
        """Return (or set) the numeric value of the variable lower bound."""
        # Note: Implementation avoids unnecessary expression generation
        domain_lb, domain_ub = self.domain.bounds()
        # lb is the tighter of the domain and bounds
        return self._resolve_bound_value(self._lb, domain_lb, _ninf, max)

    @lb.setter
    def lb(self, val):
        self.lower = val

    @property
    def ub(self):
        """Return (or set) the numeric value of the variable upper bound."""
        # Note: implementation avoids unnecessary expression generation
        domain_lb, domain_ub = self.domain.bounds()
        # ub is the tighter of the domain and bounds
        return self._resolve_bound_value(self._ub, domain_ub, _inf, min)

    @ub.setter
    def ub(self, val):
        self.upper = val

    @property
    def lower(self):
        """Return (or set) an expression for the variable lower bound.

        This returns a (not potentially variable) expression for the
        variable lower bound.  This represents the tighter of the
        current domain and the constant or expression assigned to
        :attr:`lower`.  Note that the expression will NOT automatically
        reflect changes to either the domain or the bound expression
        (e.g., because of assignment to either :attr:`lower` or
        :attr:`domain`).

        """
        dlb, _ = self.domain.bounds()
        if self._lb is None:
            return dlb
        elif dlb is None:
            return self._lb
        # _process_bound() guarantees _lb is not potentially variable
        return NPV_MaxExpression((self._lb, dlb))

    @lower.setter
    def lower(self, val):
        self._lb = self._process_bound(val, 'lower')

    @property
    def upper(self):
        """Return (or set) an expression for the variable upper bound.

        This returns a (not potentially variable) expression for the
        variable upper bound.  This represents the tighter of the
        current domain and the constant or expression assigned to
        :attr:`upper`.  Note that the expression will NOT automatically
        reflect changes to either the domain or the bound expression
        (e.g., because of assignment to either :attr:`upper` or
        :attr:`domain`).

        """
        _, dub = self.domain.bounds()
        if self._ub is None:
            return dub
        elif dub is None:
            return self._ub
        # _process_bound() guarantees _lb is not potentially variable
        return NPV_MinExpression((self._ub, dub))

    @upper.setter
    def upper(self, val):
        self._ub = self._process_bound(val, 'upper')

    def get_units(self):
        """Return the units for this variable entry."""
        # parent_component() returns self if this is scalar, or the owning
        # component if not scalar
        return self.parent_component()._units

    def fix(self, value=NOTSET, skip_validation=False):
        """Fix the value of this variable (treat as nonvariable)

        This sets the :attr:`fixed` indicator to True.  If ``value`` is
        provided, the value (and the ``skip_validation`` flag) are first
        passed to :meth:`set_value()`.

        """
        self.fixed = True
        if value is not NOTSET:
            self.set_value(value, skip_validation)

    def unfix(self):
        """Unfix this variable (treat as variable in solver interfaces)

        This sets the :attr:`fixed` indicator to False.

        """
        self.fixed = False

    def free(self):
        """Alias for :meth:`unfix`"""
        return self.unfix()

    @property
    def fixed(self):
        """Return (or set) the fixed indicator for this variable.

        Alias for :meth:`is_fixed` / :meth:`fix` / :meth:`unfix`.

        """
        return self._fixed

    @fixed.setter
    def fixed(self, val):
        self._fixed = bool(val)

    @property
    def stale(self):
        """The stale status for this variable.

        Variables are "stale" if their current value was not updated as
        part of the most recent model update.  A "model update" can be
        one of several things: a solver invocation, loading a previous
        solution, or manually updating a non-stale :class:`Var` value.

        Returns
        -------
        bool

        Notes
        -----
        Fixed :class:`Var` objects will be stale after invoking a solver
        (as their value was not updated by the solver).

        Updating a stale :class:`Var` value will not cause other
        variable values to be come stale.  However, updating the first
        non-stale :class:`Var` value after a solve or solution load
        *will* cause all other variables to be marked as stale

        """
        return StaleFlagManager.is_stale(self._stale)

    @stale.setter
    def stale(self, val):
        if val:
            self._stale = 0  # True
        else:
            self._stale = StaleFlagManager.get_flag(0)

    def is_integer(self):
        """Returns True when the domain is a contiguous integer range."""
        _id = id(self.domain)
        if _id in _known_global_real_domains:
            return not _known_global_real_domains[_id]
        _interval = self.domain.get_interval()
        if _interval is None:
            return False
        # Note: it is not sufficient to just check the step: the
        # starting / ending points must be integers (or not specified)
        start, stop, step = _interval
        return (
            step == 1
            and (start is None or int(start) == start)
            and (stop is None or int(stop) == stop)
        )

    def is_binary(self):
        """Returns True when the domain is restricted to Binary values."""
        domain = self.domain
        if domain is Binary:
            return True
        if id(domain) in _known_global_real_domains:
            return False
        return domain.get_interval() == (0, 1, 1)

    def is_continuous(self):
        """Returns True when the domain is a continuous real range"""
        _id = id(self.domain)
        if _id in _known_global_real_domains:
            return _known_global_real_domains[_id]
        _interval = self.domain.get_interval()
        return _interval is not None and _interval[2] == 0

    def is_fixed(self):
        """Returns True if this variable is fixed, otherwise returns False."""
        return self._fixed

    def is_constant(self):
        """Returns False because this is not a constant in an expression."""
        return False

    def is_variable_type(self):
        """Returns True because this is a variable."""
        return True

    def is_potentially_variable(self):
        """Returns True because this is a variable."""
        return True

    def clear(self):
        self.value = None

    def _compute_polynomial_degree(self, result):
        """
        If the variable is fixed, it represents a constant
        is a polynomial with degree 0. Otherwise, it has
        degree 1. This method is used in expressions to
        compute polynomial degree.
        """
        if self._fixed:
            return 0
        return 1

    def _process_bound(self, val, bound_type):
        if type(val) in native_numeric_types or val is None:
            # TODO: warn/error: check if this Var has units: assigning
            # a dimensionless value to a united variable should be an error
            pass
        elif is_potentially_variable(val):
            raise ValueError(
                "Potentially variable input of type '%s' supplied as "
                "%s bound for variable '%s' - legal types must be constants "
                "or non-potentially variable expressions."
                % (type(val).__name__, bound_type, self.name)
            )
        else:
            # We want to create an expression and not just convert the
            # current value so that things like mutable Params behave as
            # expected.
            _units = self.parent_component()._units
            if _units is not None:
                val = units.convert(val, to_units=_units)
        return val


class _VarData(metaclass=RenamedClass):
    __renamed__new_class__ = VarData
    __renamed__version__ = '6.7.2'


class _GeneralVarData(metaclass=RenamedClass):
    __renamed__new_class__ = VarData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register("Decision variables.")
class Var(IndexedComponent, IndexedComponent_NDArrayMixin):
    """A numeric variable, which may be defined over an index.

    Args:
        domain (Set or function, optional): A Set that defines valid
            values for the variable (e.g., ``Reals``, ``NonNegativeReals``,
            ``Binary``), or a rule that returns Sets.  Defaults to ``Reals``.
        within (Set or function, optional): An alias for ``domain``.
        bounds (tuple or function, optional): A tuple of ``(lower, upper)``
            bounds for the variable, or a rule that returns tuples.
            Defaults to ``(None, None)``.
        initialize (float or function, optional): The initial value for
            the variable, or a rule that returns initial values.
        rule (float or function, optional): An alias for ``initialize``.
        dense (bool, optional): Instantiate all elements from
            :meth:`index_set` when constructing the Var (True) or just the
            variables returned by ``initialize``/``rule`` (False).  Defaults
            to ``True``.
        units (pyomo units expression, optional): Set the units corresponding
            to the entries in this variable.
        name (str, optional): Name for this component.
        doc (str, optional): Text describing this component.
    """

    _ComponentDataClass = VarData

    @overload
    def __new__(cls: Type[ScalarVar], *args, **kwargs) -> ScalarVar: ...

    @overload
    def __new__(cls: Type[IndexedVar], *args, **kwargs) -> IndexedVar: ...

    @overload
    def __new__(cls: Type[Var], *args, **kwargs) -> Union[ScalarVar, IndexedVar]: ...

    def __new__(cls, *args, **kwargs):
        if cls is not Var:
            return super(Var, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return super(Var, cls).__new__(AbstractScalarVar)
        else:
            return super(Var, cls).__new__(IndexedVar)

    @overload
    def __init__(
        self,
        *indexes,
        domain=Reals,
        within=Reals,
        bounds=None,
        initialize=None,
        rule=None,
        dense=True,
        units=None,
        name=None,
        doc=None,
    ): ...

    def __init__(self, *args, **kwargs):
        #
        # Default keyword values
        #
        self._rule_init = Initializer(
            self._pop_from_kwargs('Var', kwargs, ('rule', 'initialize'), None)
        )
        self._rule_domain = SetInitializer(
            self._pop_from_kwargs('Var', kwargs, ('domain', 'within'), Reals)
        )
        _bounds_arg = kwargs.pop('bounds', None)
        self._dense = kwargs.pop('dense', True)
        self._units = kwargs.pop('units', None)
        if self._units is not None:
            self._units = units.get_units(self._units)
        #
        # Initialize the base class
        #
        kwargs.setdefault('ctype', Var)
        IndexedComponent.__init__(self, *args, **kwargs)
        #
        # Now that we can call is_indexed(), process bounds initializer
        #
        if not self.is_indexed() and not self._dense:
            logger.warning(
                "ScalarVar object '%s': dense=False is not allowed "
                "for scalar variables; converting to dense=True" % (self.name,)
            )
            self._dense = True
        self._rule_bounds = BoundInitializer(_bounds_arg, self)

    def flag_as_stale(self):
        """
        Set the 'stale' attribute of every variable data object to True.
        """
        for var_data in self._data.values():
            var_data.stale = True

    def get_values(self, include_fixed_values=True):
        """
        Return a dictionary of index-value pairs.
        """
        if include_fixed_values:
            return {idx: vardata.value for idx, vardata in self._data.items()}
        return {
            idx: vardata.value
            for idx, vardata in self._data.items()
            if not vardata.fixed
        }

    extract_values = get_values

    def set_values(self, new_values, skip_validation=False):
        """
        Set the values of a dictionary.

        The default behavior is to validate the values in the
        dictionary.
        """
        for index, new_value in new_values.items():
            self[index].set_value(new_value, skip_validation)

    def get_units(self):
        """Return the units expression for this Var."""
        return self._units

    # TODO: deprecate this?  __getitem__ is generally preferable"
    def add(self, index):
        """Add a variable with a particular index."""
        return self[index]

    def construct(self, data=None):
        """
        Construct the VarData objects for this variable
        """
        if self._constructed:
            return
        self._constructed = True

        timer = ConstructionTimer(self)
        if is_debug_set(logger):
            logger.debug("Constructing Variable %s" % (self.name,))

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        # Note: define 'index' to avoid 'variable referenced before
        # assignment' in the error message generated in the 'except:'
        # block below.
        index = None
        try:
            # We do not (currently) accept data for constructing Variables
            assert data is None

            if not self.index_set().isfinite() and self._dense:
                # Note: if the index is not finite, then we cannot
                # iterate over it.  This used to be fatal; now we
                # just warn
                logger.warning(
                    "Var '%s' indexed by a non-finite set, but declared "
                    "with 'dense=True'.  Reverting to 'dense=False' as "
                    "it is not possible to make this variable dense.  "
                    "This warning can be suppressed by specifying "
                    "'dense=False'" % (self.name,)
                )
                self._dense = False

            if self._rule_init is not None and self._rule_init.contains_indices():
                # Historically we have allowed Vars to be initialized by
                # a sparse map (i.e., a dict containing only some of the
                # keys).  We will wrap the incoming initializer to map
                # KeyErrors to None
                self._rule_init = DefaultInitializer(self._rule_init, None, KeyError)
                # The index is coming in externally; we need to validate it
                for index in self._rule_init.indices():
                    self[index]
                # If this is a dense object, we need to ensure that it
                # has been filled in.
                if self._dense:
                    for index in self.index_set():
                        if index not in self._data:
                            self._getitem_when_not_present(index)
            elif not self.is_indexed():
                # As there is only a single VarData to populate, just do
                # so and bypass all special-case testing below
                self._getitem_when_not_present(None)
            elif self._dense:
                # Special case: initialize every VarData.  For the
                # initializers that are constant, we can avoid
                # re-calling (and re-validating) the inputs in certain
                # cases.  To support this, we will create the first
                # VarData and then use it as a template to initialize
                # (constant portions of) every VarData so as to not
                # repeat all the domain/bounds validation.
                try:
                    ref = self._getitem_when_not_present(next(iter(self.index_set())))
                except StopIteration:
                    # Empty index!
                    return
                call_domain_rule = not self._rule_domain.constant()
                call_bounds_rule = (
                    self._rule_bounds is not None and not self._rule_bounds.constant()
                )
                call_init_rule = self._rule_init is not None and (
                    not self._rule_init.constant()
                    # If either the domain or bounds change, then we
                    # need to re-verify the initial value, even if it is
                    # constant:
                    or call_domain_rule
                    or call_bounds_rule
                )
                # Initialize all the component datas with the common data
                for index in self.index_set():
                    self._data[index] = self._ComponentDataClass.copy(ref)
                    # NOTE: This is a special case where a key, value pair is
                    # added to the _data dictionary without calling
                    # _getitem_when_not_present, which is why we need to set the
                    # index here.
                    self._data[index]._index = index
                # Now go back and initialize any index-specific data
                block = self.parent_block()
                if call_domain_rule:
                    for index, obj in self._data.items():
                        # We can directly set the attribute (not the
                        # property) because the SetInitializer ensures
                        # that the value is a proper Set.
                        obj._domain = self._rule_domain(block, index, self)
                if call_bounds_rule:
                    for index, obj in self._data.items():
                        obj.lower, obj.upper = self._rule_bounds(block, index)
                if call_init_rule:
                    for index, obj in self._data.items():
                        obj.set_value(self._rule_init(block, index))
            else:
                # non-dense indexed var with generic
                # (non-index-containing) initializer: nothing to do
                pass

        except Exception:
            err = sys.exc_info()[1]
            logger.error(
                "Rule failed when initializing variable for "
                "Var %s with index %s:\n%s: %s"
                % (self.name, str(index), type(err).__name__, err)
            )
            raise
        finally:
            timer.report()

    #
    # This method must be defined on subclasses of
    # IndexedComponent that support implicit definition
    #
    def _getitem_when_not_present(self, index):
        """Returns the default component data value."""
        if index is None and not self.is_indexed():
            obj = self._data[index] = self
        else:
            obj = self._data[index] = self._ComponentDataClass(component=self)
        parent = self.parent_block()
        obj._index = index
        # We can directly set the attribute (not the property) because
        # the SetInitializer ensures that the value is a proper Set.
        obj._domain = self._rule_domain(parent, index, self)
        if self._rule_bounds is not None:
            obj.lower, obj.upper = self._rule_bounds(parent, index)
        if self._rule_init is not None:
            obj.set_value(self._rule_init(parent, index))
        return obj

    #
    # Because we need to do more initialization than simply calling
    # set_value(), we need to override _setitem_when_not_present
    #
    def _setitem_when_not_present(self, index, value=NOTSET):
        if value is self.Skip:
            return None
        try:
            obj = self._getitem_when_not_present(index)
            if value is not NOTSET:
                obj.set_value(value)
        except:
            self._data.pop(index, None)
            raise
        return obj

    def _pprint(self):
        """Print component information."""
        headers = [
            ("Size", len(self)),
            ("Index", self._index_set if self.is_indexed() else None),
        ]
        if self._units is not None:
            headers.append(('Units', str(self._units)))
        return (
            headers,
            self.items,
            ("Lower", "Value", "Upper", "Fixed", "Stale", "Domain"),
            lambda k, v: [
                value(v.lb),
                v.value,
                value(v.ub),
                v.fixed,
                v.stale,
                v.domain,
            ],
        )


class ScalarVar(VarData, Var):
    """A single variable."""

    def __init__(self, *args, **kwd):
        VarData.__init__(self, component=self)
        Var.__init__(self, *args, **kwd)
        self._index = UnindexedComponent_index


@disable_methods(_VARDATA_API)
class AbstractScalarVar(ScalarVar):
    pass


class SimpleVar(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarVar
    __renamed__version__ = '6.0'


class IndexedVar(Var):
    """An array of variables."""

    def setlb(self, val):
        """
        Set the lower bound for this variable.
        """
        for vardata in self.values():
            vardata.lower = val

    def setub(self, val):
        """
        Set the upper bound for this variable.
        """
        for vardata in self.values():
            vardata.upper = val

    def fix(self, value=NOTSET, skip_validation=False):
        """Fix all variables in this :class:`IndexedVar` (treat as nonvariable)

        This sets the :attr:`fixed` indicator to True for every variable
        in this IndexedVar.  If ``value`` is provided, the value (and
        the ``skip_validation`` flag) are first passed to
        :meth:`set_value`.

        """
        for vardata in self.values():
            vardata.fix(value, skip_validation)

    def unfix(self):
        """Unfix all variables in this :class:`IndexedVar` (treat as variable)

        This sets the :attr:`VarData.fixed` indicator to False for
        every variable in this :class:`IndexedVar`.

        """
        for vardata in self.values():
            vardata.unfix()

    def free(self):
        """Alias for :meth:`unfix`"""
        return self.unfix()

    @property
    def domain(self):
        raise AttributeError(
            "The domain is not an attribute for IndexedVar. It "
            "can be set for all indices using this property setter, "
            "but must be accessed for individual variables in this container."
        )

    @domain.setter
    def domain(self, domain):
        """Sets the domain for all variables in this container."""
        try:
            domain_rule = SetInitializer(domain)
            if domain_rule.constant():
                domain = domain_rule(self.parent_block(), None, self)
                for vardata in self.values():
                    vardata._domain = domain
            elif domain_rule.contains_indices():
                parent = self.parent_block()
                for index in domain_rule.indices():
                    self[index]._domain = domain_rule(parent, index, self)
            else:
                parent = self.parent_block()
                for index, vardata in self.items():
                    vardata._domain = domain_rule(parent, index, self)
        except:
            logger.error(
                "%s is not a valid domain. Variable domains must be an "
                "instance of a Pyomo Set or convertible to a Pyomo Set." % (domain,),
                extra={'id': 'E2001'},
            )
            raise

    # Because CP supports indirection [the ability to index objects by
    # another (inter) Var] for certain types (including Var), we will
    # catch the normal RuntimeError and return a (variable)
    # GetItemExpression.
    #
    # FIXME: We should integrate this logic into the base implementation
    # of `__getitem__()`, including the recognition / differentiation
    # between potentially variable GetItemExpression objects and
    # "constant" GetItemExpression objects.  That will need to wait for
    # the expression rework [JDS; Nov 22].
    def __getitem__(self, args) -> VarData:
        try:
            return super().__getitem__(args)
        except RuntimeError:
            tmp = args if args.__class__ is tuple else (args,)
            if any(
                hasattr(arg, 'is_potentially_variable')
                and arg.is_potentially_variable()
                for arg in tmp
            ):
                return GetItemExpression((self,) + tmp)
            raise


@ModelComponentFactory.register("List of decision variables.")
class VarList(IndexedVar):
    """
    Variable-length indexed variable objects used to construct Pyomo models.
    """

    def __init__(self, **kwargs):
        self._starting_index = kwargs.pop('starting_index', 1)
        args = (Set(dimen=1),)
        IndexedVar.__init__(self, *args, **kwargs)

    def construct(self, data=None):
        """Construct this component."""
        if self._constructed:
            return
        # Note: do not set _constructed here, or the super() call will
        # not actually construct the component.

        if is_debug_set(logger):
            logger.debug("Constructing variable list %s", self.name)

        self.index_set().construct()

        # We need to ensure that the indices needed for initialization are
        # added to the underlying implicit set.  We *could* verify that the
        # indices in the initialization dict are all sequential integers,
        # OR we can just add the correct number of sequential integers and
        # then let _validate_index complain when we set the value.
        if self._rule_init is not None and self._rule_init.contains_indices():
            for i, idx in enumerate(self._rule_init.indices()):
                self._index_set.add(i + self._starting_index)
        super(VarList, self).construct(data)

    def add(self):
        """Add a variable to this list."""
        next_idx = len(self._index_set) + self._starting_index
        self._index_set.add(next_idx)
        return self[next_idx]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Tests for the MindtPy solver."""

import pyomo.common.unittest as unittest
from pyomo.contrib.mindtpy.tests.eight_process_problem import EightProcessFlowsheet
from pyomo.contrib.mindtpy.tests.nonconvex1 import Nonconvex1
from pyomo.contrib.mindtpy.tests.nonconvex2 import Nonconvex2
from pyomo.contrib.mindtpy.tests.nonconvex3 import Nonconvex3
from pyomo.contrib.mindtpy.tests.nonconvex4 import Nonconvex4
from pyomo.environ import SolverFactory, value
from pyomo.environ import *
from pyomo.opt import TerminationCondition

required_solvers = ('baron', 'cplex_persistent')
if not all(SolverFactory(s).available(exception_flag=False) for s in required_solvers):
    subsolvers_available = False
elif not SolverFactory('baron').license_is_valid():
    subsolvers_available = False
else:
    subsolvers_available = True

model_list = [
    EightProcessFlowsheet(convex=False),
    Nonconvex1(),
    Nonconvex2(),
    Nonconvex3(),
    Nonconvex4(),
]


@unittest.skipIf(
    not subsolvers_available,
    'Required subsolvers %s are not available' % (required_solvers,),
)
@unittest.skipIf(
    not pyomo.contrib.mcpp.pyomo_mcpp.mcpp_available(), 'MC++ is not available'
)
class TestMindtPy(unittest.TestCase):
    """Tests for the MindtPy solver plugin."""

    def check_optimal_solution(self, model, places=1):
        for var in model.optimal_solution:
            self.assertAlmostEqual(
                var.value, model.optimal_solution[var], places=places
            )

    def test_GOA(self):
        """Test the global outer approximation decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='GOA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=2
                )
                self.check_optimal_solution(model)

    def test_GOA_tabu_list(self):
        """Test the global outer approximation decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='GOA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    use_tabu_list=True,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=2
                )
                self.check_optimal_solution(model)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.errors import (
    format_exception,
    PyomoException,
    DeveloperError,
    MouseTrap,
)


class LocalException(Exception):
    pass


class CustomLocalException(PyomoException):
    default_message = 'Default message.'


class TestFormatException(unittest.TestCase):
    def test_basic_message(self):
        self.assertEqual(format_exception("Hello world"), "Hello world")

    def test_formatted_message(self):
        self.assertEqual(format_exception("Hello\nworld"), "Hello\nworld")

    def test_long_basic_message(self):
        self.assertEqual(
            format_exception(
                "Hello world, this is a very long message that will "
                "inevitably wrap onto another line."
            ),
            "Hello world, this is a very long message that will\n"
            "    inevitably wrap onto another line.",
        )

    def test_long_basic_message_exception(self):
        self.assertEqual(
            format_exception(
                "Hello world, this is a very long message that will "
                "inevitably wrap onto another line.",
                exception=LocalException(),
            ),
            "Hello world, this is a very\n"
            "    long message that will inevitably wrap onto another line.",
        )

    def test_long_basic_message_builtin_exception(self):
        self.assertEqual(
            format_exception(
                "Hello world, this is a very long message that will "
                "inevitably wrap onto another line.",
                exception=RuntimeError,
            ),
            "Hello world, this is a very long message that will inevitably\n"
            "    wrap onto another line.",
        )

    def test_basic_message_prolog(self):
        self.assertEqual(
            format_exception(
                "This is a very, very, very long message that will "
                "inevitably wrap onto another line.",
                prolog="Hello world:",
            ),
            "Hello world:\n"
            "    This is a very, very, very long message that will inevitably "
            "wrap onto\n"
            "    another line.",
        )

    def test_basic_message_long_prolog(self):
        msg = format_exception(
            "This is a very, very, very long message that will "
            "inevitably wrap onto another line.",
            prolog="Hello, this is a more verbose prolog that will "
            "trigger a line wrap:",
        )
        self.assertEqual(
            msg,
            "Hello, this is a more verbose prolog that will trigger\n"
            "    a line wrap:\n"
            "        This is a very, very, very long message that will inevitably "
            "wrap\n"
            "        onto another line.",
        )

    def test_basic_message_formatted_prolog(self):
        msg = format_exception(
            "This is a very, very, very long message that will "
            "inevitably wrap onto another line.",
            prolog="Hello world:\n    This is a prolog:",
        )
        self.assertEqual(
            msg,
            "Hello world:\n    This is a prolog:\n"
            "        This is a very, very, very long message that will inevitably "
            "wrap\n"
            "        onto another line.",
        )

    def test_basic_message_epilog(self):
        self.assertEqual(
            format_exception(
                "This is a very, very, very long message that will "
                "inevitably wrap onto another line.",
                epilog="Hello world",
            ),
            "This is a very, very, very long message that will\n"
            "        inevitably wrap onto another line.\n"
            "    Hello world",
        )

    def test_basic_message_long_epilog(self):
        self.assertEqual(
            format_exception(
                "This is a very, very, very long message that will "
                "inevitably wrap onto another line.",
                epilog="Hello, this is a very, very, very verbose epilog that will "
                "trigger a line wrap",
            ),
            "This is a very, very, very long message that will\n"
            "        inevitably wrap onto another line.\n"
            "    Hello, this is a very, very, very verbose epilog that will trigger a\n"
            "    line wrap",
        )

    def test_basic_message_formatted_epilog(self):
        msg = format_exception(
            "This is a very, very, very long message that will "
            "inevitably wrap onto another line.",
            epilog="Hello world:\n    This is an epilog:",
        )
        self.assertEqual(
            msg,
            "This is a very, very, very long message that will\n"
            "        inevitably wrap onto another line.\n"
            "Hello world:\n    This is an epilog:",
        )

    def test_format_empty_message(self):
        msg = format_exception('')
        self.assertEqual('', msg)

        msg = format_exception(repr(''))
        self.assertEqual("''", msg)

        msg = format_exception('', prolog="dev error:")
        self.assertEqual("dev error:", msg)

        msg = format_exception(repr(''), prolog="dev error:")
        self.assertEqual("dev error:\n    ''", msg)


class TestPyomoException(unittest.TestCase):
    def test_default_message(self):
        exception = CustomLocalException()
        self.assertIn("Default message.", str(exception))

    def test_custom_message_override(self):
        exception = CustomLocalException("Non-default message.")
        self.assertNotIn("Default message.", str(exception))
        self.assertIn("Non-default message.", str(exception))

    def test_DeveloperError(self):
        self.assertEqual(
            "Internal Pyomo implementation error:\n"
            "    Please report this to the Pyomo Developers.",
            str(DeveloperError()),
        )

        self.assertEqual(
            "Internal Pyomo implementation error:\n"
            "        'Error message'\n"
            "    Please report this to the Pyomo Developers.",
            str(DeveloperError("Error message")),
        )

    def test_MouseTrap(self):
        self.assertEqual(
            """Sorry, mouse, no cookies here!
        ''
    This is functionality we think may be rational to support, but is not
    yet implemented (possibly due to developer availability, complexity of
    edge cases, or general practicality or tractability). However, please
    feed the mice: pull requests are always welcome!""",
            str(MouseTrap()),
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Utility functions
#
import inspect

from pyomo.common.deprecation import relocated_module_attribute
from pyomo.core.base.indexed_component import normalize_index

relocated_module_attribute(
    'disable_methods', 'pyomo.core.base.disable_methods.disable_methods', version='6.1'
)
relocated_module_attribute(
    'Initializer', 'pyomo.core.base.initializer.Initializer', version='6.1'
)
relocated_module_attribute(
    'IndexedCallInitializer', 'pyomo.core.base.initializer.Initializer', version='6.1'
)
relocated_module_attribute(
    'CountedCallInitializer', 'pyomo.core.base.initializer.Initializer', version='6.1'
)


def is_functor(obj):
    """
    Returns true iff obj.__call__ is defined.
    """
    return inspect.isfunction(obj) or hasattr(obj, '__call__')


def flatten_tuple(x):
    """
    This wraps around normalize_index. It flattens a nested sequence into
    a single tuple and always returns a tuple, even for single
    element inputs.

    Returns
    -------
    tuple

    """
    x = normalize_index(x)
    if isinstance(x, tuple):
        return x
    return (x,)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.opt.base import SolverFactory, ProblemFormat, ResultsFormat
from pyomo.opt.solver import SystemCallSolver
from pyomo.common.collections import Bunch
from pyomo.common.tempfiles import TempfileManager

logger = logging.getLogger('pyomo.neos')


@SolverFactory.register('_neos', 'Interface for solvers hosted on NEOS')
class NEOSRemoteSolver(SystemCallSolver):
    """A wrapper class for NEOS Remote Solvers"""

    def __init__(self, **kwds):
        kwds["type"] = "neos"
        SystemCallSolver.__init__(self, **kwds)
        self._valid_problem_formats = [ProblemFormat.nl]
        self._valid_result_formats = {}
        self._valid_result_formats[ProblemFormat.nl] = [ResultsFormat.sol]
        self._problem_format = ProblemFormat.nl
        self._results_format = ResultsFormat.sol

    def create_command_line(self, executable, problem_files):
        """
        Create the local ``*.sol`` and ``*.log`` files, which will be
        populated by NEOS.
        """
        if self._log_file is None:
            self._log_file = TempfileManager.create_tempfile(suffix=".neos.log")
        if self._soln_file is None:
            self._soln_file = TempfileManager.create_tempfile(suffix=".neos.sol")
            self._results_file = self._soln_file

        # display the log/solver file names prior to execution. this is useful
        # in case something crashes unexpectedly, which is not without precedent.
        if self._keepfiles:
            if self._log_file is not None:
                logger.info("Solver log file: '%s'" % (self._log_file,))
            if self._soln_file is not None:
                logger.info("Solver solution file: '%s'" % (self._soln_file,))
            if self._problem_files != []:
                logger.info("Solver problem files: %s" % (self._problem_files,))

        return Bunch(cmd="", log_file=self._log_file, env="")

    def _default_executable(self):
        return True



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import namedtuple
import enum
import itertools as it
import traceback
from pyomo.common.config import document_kwargs_from_configdict
from pyomo.common.errors import InfeasibleConstraintException
from pyomo.contrib.fbbt.fbbt import fbbt
from pyomo.contrib.gdpopt.algorithm_base_class import _GDPoptAlgorithm
from pyomo.contrib.gdpopt.create_oa_subproblems import (
    add_util_block,
    add_disjunction_list,
    add_disjunct_list,
    add_algebraic_variable_list,
    add_boolean_variable_lists,
    add_transformed_boolean_variable_list,
)
from pyomo.contrib.gdpopt.config_options import (
    _add_nlp_solver_configs,
    _add_ldsda_configs,
    _add_mip_solver_configs,
    _add_tolerance_configs,
    _add_nlp_solve_configs,
)
from pyomo.contrib.gdpopt.nlp_initialization import restore_vars_to_original_values
from pyomo.contrib.gdpopt.util import SuppressInfeasibleWarning, get_main_elapsed_time
from pyomo.contrib.satsolver.satsolver import satisfiable
from pyomo.core import minimize, Suffix, TransformationFactory, Objective, value
from pyomo.opt import SolverFactory
from pyomo.opt import TerminationCondition as tc
from pyomo.core.expr.logical_expr import ExactlyExpression


class DirectionNorm(str, enum.Enum):
    """
    Norm type for search direction generation in LD-SDA.

    Attributes
    ----------
    L2 : str
        Standard basis vectors (2n directions for n external variables).
    Linf : str
        All combinations of {-1, 0, 1} excluding the zero vector (3^n - 1 directions).
    """

    L2 = 'L2'
    Linf = 'Linf'

    def __str__(self):
        return self.value


class SearchPhase(str, enum.Enum):
    """
    Phase of the LD-SDA search algorithm.

    Attributes
    ----------
    INITIAL : str
        Initial point evaluation.
    NEIGHBOR : str
        Neighbor search phase.
    LINE : str
        Line search phase.
    """

    INITIAL = 'Initial point'
    NEIGHBOR = 'Neighbor search'
    LINE = 'Line search'

    def __str__(self):
        return self.value


# Data tuple for external variables.
ExternalVarInfo = namedtuple(
    'ExternalVarInfo',
    [
        'exactly_number',  # number of external variables for this type
        'Boolean_vars',  # list with names of the ordered Boolean variables to be reformulated
        'UB',  # upper bound on external variable
        'LB',  # lower bound on external variable
    ],
)


@SolverFactory.register(
    'gdpopt.ldsda',
    doc="The LD-SDA (Logic-based Discrete-Steepest Descent Algorithm) "
    "Generalized Disjunctive Programming (GDP) solver",
)
class GDP_LDSDA_Solver(_GDPoptAlgorithm):
    """
    The GDPopt (Generalized Disjunctive Programming optimizer) LD-SDA
    (Logic-based Discrete-Steepest Descent Algorithm) solver.

    This solver accepts models that can include nonlinear, continuous variables
    and constraints, as well as logical conditions. It uses a discrete steepest
    descent approach to explore the space of discrete variables (disjunctions)
    while solving NLP subproblems for the continuous variables.

    References
    ----------
    Ovalle, D.; Liñán, D. A.; Lee, A.; Gómez, J. M.; Ricardez-Sandoval, L.; Grossmann, I. E.; Bernal Neira, D. E. Logic-Based Discrete-Steepest Descent: A Solution Method for Process Synthesis Generalized Disjunctive Programs. Computers & Chemical Engineering 2025, 195, 108993. https://doi.org/10.1016/j.compchemeng.2024.108993.
    """

    CONFIG = _GDPoptAlgorithm.CONFIG()
    _add_mip_solver_configs(CONFIG)
    _add_nlp_solver_configs(CONFIG, default_solver='ipopt')
    _add_nlp_solve_configs(
        CONFIG, default_nlp_init_method=restore_vars_to_original_values
    )
    _add_tolerance_configs(CONFIG)
    _add_ldsda_configs(CONFIG)

    algorithm = 'LDSDA'

    # Override solve() to customize the docstring for this solver
    @document_kwargs_from_configdict(CONFIG, doc=_GDPoptAlgorithm.solve.__doc__)
    def solve(self, model, **kwds):
        return super().solve(model, **kwds)

    def _log_citation(self, config):
        config.logger.info("\n" + """- LDSDA algorithm:
        Bernal DE, Ovalle D, Liñán DA, Ricardez-Sandoval LA, Gómez JM, Grossmann IE.
        Process Superstructure Optimization through Discrete Steepest Descent Optimization: a GDP Analysis and Applications in Process Intensification.
        Computer Aided Chemical Engineering 2022 Jan 1 (Vol. 49, pp. 1279-1284). Elsevier.
        https://doi.org/10.1016/B978-0-323-85159-6.50213-X
        """.strip())

    def _solve_gdp(self, model, config):
        """
        Execute the main LD-SDA algorithm logic.

        Initializes the utility blocks, reformulates the model, solves the
        initial point, and enters the main search loop (Neighbor Search and
        Line Search) until a local optimum is found or termination criteria
        are met.

        Parameters
        ----------
        model : ConcreteModel
            The GDP model to be solved.
        config : ConfigBlock
            The configuration block containing solver options.
        """
        logger = config.logger
        self.log_formatter = (
            '{:>9}   {:>15}   {:>20}   {:>11.5f}   {:>11.5f}   {:>8.2%}   {:>7.2f}  {}'
        )
        self.best_direction = None
        self.current_point = tuple(config.starting_point)
        self.explored_point_set = set()

        # Create utility block on the original model so that we will be able to
        # copy solutions between
        util_block = self.original_util_block = add_util_block(model)
        add_disjunct_list(util_block)
        add_algebraic_variable_list(util_block)
        add_boolean_variable_lists(util_block)
        util_block.config_disjunction_list = config.disjunction_list
        util_block.config_logical_constraint_list = config.logical_constraint_list

        # We will use the working_model to perform the LDSDA search.
        self.working_model = model.clone()
        self.working_model_util_block = self.working_model.find_component(util_block)

        add_disjunction_list(self.working_model_util_block)
        TransformationFactory('core.logical_to_linear').apply_to(self.working_model)
        # Now that logical_to_disjunctive has been called.
        add_transformed_boolean_variable_list(self.working_model_util_block)
        self.get_external_information(self.working_model_util_block, config)
        self.directions = self._get_directions(
            self.number_of_external_variables, config
        )

        # Add the BigM suffix if it does not already exist. Used later during
        # nonlinear constraint activation.
        if not hasattr(self.working_model_util_block, 'BigM'):
            self.working_model_util_block.BigM = Suffix()
        self._log_header(logger)
        # Solve the initial point
        _ = self._solve_GDP_subproblem(self.current_point, SearchPhase.INITIAL, config)

        # Main loop
        locally_optimal = False
        while not locally_optimal:
            self.iteration += 1
            if self.any_termination_criterion_met(config):
                break
            locally_optimal = self.neighbor_search(config)
            if not locally_optimal:
                self.line_search(config)

    def any_termination_criterion_met(self, config):
        """
        Check if any termination criteria (iteration limit or time limit) have been met.

        Parameters
        ----------
        config : ConfigBlock
            The configuration block containing limits.

        Returns
        -------
        bool
            True if either the iteration limit or time limit has been reached,
            False otherwise.
        """
        return self.reached_iteration_limit(config) or self.reached_time_limit(config)

    def _solve_GDP_subproblem(self, external_var_value, search_type, config):
        """
        Solve the GDP subproblem with disjunctions fixed according to the external variable values.

        This method fixes the Boolean variables based on the `external_var_value`,
        applies necessary transformations (BigM, FBBT), and solves the resulting
        MINLP/NLP using the configured solver.

        Parameters
        ----------
        external_var_value : tuple or list
            The values of the external variables (indices of active disjuncts)
            defining the current point in the discrete space.
        search_type : SearchPhase
            The context of the solve (SearchPhase.INITIAL, SearchPhase.NEIGHBOR, or SearchPhase.LINE).
        config : ConfigBlock
            The configuration block containing solver options.

        Returns
        -------
        primal_improved : bool
            True if the solution of this subproblem improved the best known
            primal bound (incumbent).
        primal_bound : float
            The objective value (primal bound) obtained from the subproblem.
            Returns None if the subproblem was infeasible.
        """
        self.fix_disjunctions_with_external_var(external_var_value)
        subproblem = self.working_model.clone()
        TransformationFactory('core.logical_to_linear').apply_to(subproblem)

        with SuppressInfeasibleWarning():
            try:
                TransformationFactory('gdp.bigm').apply_to(subproblem)
                fbbt(subproblem, integer_tol=config.integer_tolerance)
                TransformationFactory('contrib.detect_fixed_vars').apply_to(subproblem)
                TransformationFactory('contrib.propagate_fixed_vars').apply_to(
                    subproblem
                )
                TransformationFactory(
                    'contrib.deactivate_trivial_constraints'
                ).apply_to(subproblem, tmp=False, ignore_infeasible=False)
            except InfeasibleConstraintException:
                return False, None
            minlp_args = dict(config.minlp_solver_args)
            if config.time_limit is not None and config.minlp_solver == 'gams':
                elapsed = get_main_elapsed_time(self.timing)
                remaining = max(config.time_limit - elapsed, 1)
                minlp_args['add_options'] = minlp_args.get('add_options', [])
                minlp_args['add_options'].append('option reslim=%s;' % remaining)
            result = SolverFactory(config.minlp_solver).solve(subproblem, **minlp_args)
            primal_improved = self._handle_subproblem_result(
                result, subproblem, external_var_value, config, search_type
            )
            # Only retrieve primal_bound if the solve succeeded; otherwise return None
            if primal_improved:
                obj = next(subproblem.component_data_objects(Objective, active=True))
                primal_bound = value(obj)
            else:
                primal_bound = None
        return primal_improved, primal_bound

    def get_external_information(self, util_block, config):
        """
        Extract information from the model to perform the reformulation with external variables.

        Identifies logical constraints (specifically `ExactlyExpression`) or
        disjunctions to map them to external integer variables used for the
        discrete search.

        Parameters
        ----------
        util_block : Block
            The GDPopt utility block of the model where metadata is stored.
        config : ConfigBlock
            The configuration block containing logical constraint or disjunction lists.

        Raises
        ------
        ValueError
            If a logical constraint is not an `ExactlyExpression`.
            If an `Exactly(N)` constraint has N > 1.
            If the length of the starting point does not match the number of
            external variables derived.
        """
        util_block.external_var_info_list = []
        model = util_block.parent_block()
        reformulation_summary = []
        # Identify the variables that can be reformulated by performing a loop over logical constraints
        # TODO: we can automatically find all Exactly logical constraints in the model.
        # However, we cannot link the starting point and the logical constraint.
        # for c in util_block.logical_constraint_list:
        #     if isinstance(c.body, ExactlyExpression):
        if config.logical_constraint_list is not None:
            for c in util_block.config_logical_constraint_list:
                if not isinstance(c.body, ExactlyExpression):
                    raise ValueError(
                        "The logical_constraint_list config should be a list of ExactlyExpression logical constraints."
                    )
                # TODO: in the first version, we don't support more than one exactly constraint.
                exactly_number = c.body.args[0]
                if exactly_number > 1:
                    raise ValueError("The function only works for exactly_number = 1")
                sorted_boolean_var_list = c.body.args[1:]
                util_block.external_var_info_list.append(
                    ExternalVarInfo(
                        exactly_number=1,
                        Boolean_vars=sorted_boolean_var_list,
                        UB=len(sorted_boolean_var_list),
                        LB=1,
                    )
                )
                reformulation_summary.append(
                    [
                        1,
                        len(sorted_boolean_var_list),
                        [boolean_var.name for boolean_var in sorted_boolean_var_list],
                    ]
                )
        if config.disjunction_list is not None:
            for disjunction in util_block.config_disjunction_list:
                sorted_boolean_var_list = [
                    disjunct.indicator_var for disjunct in disjunction.disjuncts
                ]
                util_block.external_var_info_list.append(
                    ExternalVarInfo(
                        exactly_number=1,
                        Boolean_vars=sorted_boolean_var_list,
                        UB=len(sorted_boolean_var_list),
                        LB=1,
                    )
                )
                reformulation_summary.append(
                    [
                        1,
                        len(sorted_boolean_var_list),
                        [boolean_var.name for boolean_var in sorted_boolean_var_list],
                    ]
                )
        config.logger.info("Reformulation Summary:")
        config.logger.info("  Index | Ext Var | LB | UB | Associated Boolean Vars")
        for idx, row in enumerate(reformulation_summary):
            config.logger.info(f"  {idx} | {row[0]} | {row[1]} | {row[2]}")
        self.number_of_external_variables = sum(
            external_var_info.exactly_number
            for external_var_info in util_block.external_var_info_list
        )
        if self.number_of_external_variables != len(config.starting_point):
            raise ValueError(
                "The length of the provided starting point doesn't equal the number of disjunctions."
            )

    def fix_disjunctions_with_external_var(self, external_var_values_list):
        """
        Fix the disjunctions in the working model based on external variable values.

        Maps the integer values in `external_var_values_list` to the corresponding
        Boolean variables in the model, fixing the selected one to True and
        others to False for each logical group.

        Parameters
        ----------
        external_var_values_list : list or tuple
            The list of integer values representing the active disjunct index
            for each external variable.
        """
        for external_variable_value, external_var_info in zip(
            external_var_values_list,
            self.working_model_util_block.external_var_info_list,
        ):
            for idx, boolean_var in enumerate(external_var_info.Boolean_vars):
                if idx == external_variable_value - 1:
                    boolean_var.fix(True)
                    if boolean_var.get_associated_binary() is not None:
                        boolean_var.get_associated_binary().fix(1)
                else:
                    boolean_var.fix(False)
                    if boolean_var.get_associated_binary() is not None:
                        boolean_var.get_associated_binary().fix(0)
        self.explored_point_set.add(tuple(external_var_values_list))

    def _get_directions(self, dimension, config):
        """
        Generate the search directions for the given dimension.

        Parameters
        ----------
        dimension : int
            The dimensionality of the neighborhood (number of external variables).
        config : ConfigBlock
            The configuration block specifying the norm ('L2' or 'Linf').

        Returns
        -------
        list of tuple
            A list of direction vectors (tuples).
            - If DirectionNorm.L2: Standard basis vectors and their negatives.
            - If DirectionNorm.Linf: All combinations of {-1, 0, 1} excluding the zero vector.
        """
        if config.direction_norm == DirectionNorm.L2:
            directions = []
            for i in range(dimension):
                directions.append(tuple([0] * i + [1] + [0] * (dimension - i - 1)))
                directions.append(tuple([0] * i + [-1] + [0] * (dimension - i - 1)))
            return directions
        elif config.direction_norm == DirectionNorm.Linf:
            directions = list(it.product([-1, 0, 1], repeat=dimension))
            directions.remove((0,) * dimension)  # Remove the zero direction
            return directions

    def _check_valid_neighbor(self, neighbor):
        """
        Check if a given neighbor point is valid.

        A neighbor is valid if it has not been explored yet and lies within
        the defined bounds (LB and UB) of the external variables.

        Parameters
        ----------
        neighbor : tuple
            The coordinates of the neighbor point to check.

        Returns
        -------
        bool
            True if the neighbor is valid (unexplored and within bounds),
            False otherwise.
        """
        if neighbor in self.explored_point_set:
            return False
        return all(
            external_var_value >= external_var_info.LB
            and external_var_value <= external_var_info.UB
            for external_var_value, external_var_info in zip(
                neighbor, self.working_model_util_block.external_var_info_list
            )
        )

    def neighbor_search(self, config):
        """
        Evaluate immediate neighbors of the current point to find a better solution.

        Iterates through all search directions, generates neighbors, and solves
        their subproblems. Uses a tie-breaking mechanism favoring points farther
        away (Euclidean distance) if objective values are within tolerance.

        Parameters
        ----------
        config : ConfigBlock
            The configuration block containing solver options.

        Returns
        -------
        bool
            True if the current point is locally optimal (no better neighbor found),
            False if a better neighbor was found (current point updated).
        """
        locally_optimal = True
        best_neighbor = None
        self.best_direction = None  # reset best direction
        fmin = float('inf')  # Initialize the best objective value
        best_dist = 0  # Initialize the best distance
        abs_tol = (
            config.integer_tolerance
        )  # Use integer_tolerance for objective comparison

        # Loop through all possible directions (neighbors)
        for direction in self.directions:
            # Generate a neighbor point by applying the direction to the current point
            neighbor = tuple(map(sum, zip(self.current_point, direction)))

            # Check if the neighbor is valid
            if self._check_valid_neighbor(neighbor):
                # Solve the subproblem for this neighbor
                primal_improved, primal_bound = self._solve_GDP_subproblem(
                    neighbor, SearchPhase.NEIGHBOR, config
                )

                if primal_improved:
                    locally_optimal = False

                    # --- Tiebreaker Logic ---
                    if abs(fmin - primal_bound) < abs_tol:
                        # Calculate the squared Euclidean distance from the current point
                        dist = sum(
                            (x - y) ** 2 for x, y in zip(neighbor, self.current_point)
                        )

                        # Update the best neighbor if this one is farther away
                        if dist > best_dist:
                            best_neighbor = neighbor
                            self.best_direction = direction
                            best_dist = dist  # Update the best distance
                    else:
                        # Standard improvement logic: update if the objective is better
                        fmin = primal_bound  # Update the best objective value
                        best_neighbor = neighbor  # Update the best neighbor
                        self.best_direction = direction  # Update the best direction
                        best_dist = sum(
                            (x - y) ** 2 for x, y in zip(neighbor, self.current_point)
                        )
                    # --- End of Tiebreaker Logic ---

        # Move to the best neighbor if an improvement was found
        if not locally_optimal:
            self.current_point = best_neighbor

        return locally_optimal

    def line_search(self, config):
        """
        Perform a line search along the best direction found by the neighbor search.

        Continues moving in `self.best_direction` as long as the objective
        function value improves.

        Parameters
        ----------
        config : ConfigBlock
            The configuration block containing solver options.
        """
        primal_improved = True
        while primal_improved:
            next_point = tuple(map(sum, zip(self.current_point, self.best_direction)))
            if self._check_valid_neighbor(next_point):
                # Unpack the tuple and use only the first boolean value
                primal_improved, _ = self._solve_GDP_subproblem(
                    next_point, SearchPhase.LINE, config
                )
                if primal_improved:
                    self.current_point = next_point
                else:
                    break
            else:
                break

    def _handle_subproblem_result(
        self, subproblem_result, subproblem, external_var_value, config, search_type
    ):
        """
        Process the result of a subproblem solve.

        Checks termination conditions, updates primal bounds if valid, and
        logs the state.

        Parameters
        ----------
        subproblem_result : SolverResults
            The result object returned by the solver.
        subproblem : ConcreteModel
            The subproblem model instance.
        external_var_value : tuple
            The external variable configuration used for this subproblem.
        config : ConfigBlock
            The configuration block.
        search_type : SearchPhase
            The type of search (SearchPhase.NEIGHBOR, etc.).

        Returns
        -------
        bool
            True if the result improved the current best primal bound,
            False otherwise.
        """
        if subproblem_result is None:
            return False
        if subproblem_result.solver.termination_condition in {
            tc.optimal,
            tc.feasible,
            tc.globallyOptimal,
            tc.locallyOptimal,
            tc.maxTimeLimit,
            tc.maxIterations,
            tc.maxEvaluations,
        }:
            primal_bound = (
                subproblem_result.problem.upper_bound
                if self.objective_sense == minimize
                else subproblem_result.problem.lower_bound
            )
            primal_improved = self._update_bounds_after_solve(
                search_type,
                primal=primal_bound,
                logger=config.logger,
                current_point=external_var_value,
            )
            if primal_improved:
                self.update_incumbent(
                    subproblem.component(self.original_util_block.name)
                )
            return primal_improved
        return False

    def _log_header(self, logger):
        logger.info(
            '================================================================='
            '===================================='
        )
        logger.info(
            '{:^9} | {:^15} | {:^20} | {:^11} | {:^11} | {:^8} | {:^7}\n'.format(
                'Iteration',
                'Search Type',
                'External Variables',
                'Lower Bound',
                'Upper Bound',
                'Gap',
                'Time(s)',
            )
        )

    def _log_current_state(
        self, logger, search_type, current_point, primal_improved=False
    ):
        star = "*" if primal_improved else ""
        logger.info(
            self.log_formatter.format(
                self.iteration,
                search_type,
                str(current_point),
                self.LB,
                self.UB,
                self.relative_gap(),
                get_main_elapsed_time(self.timing),
                star,
            )
        )

    def _update_bounds_after_solve(
        self, search_type, primal=None, dual=None, logger=None, current_point=None
    ):
        primal_improved = self._update_bounds(primal, dual)
        if logger is not None:
            self._log_current_state(logger, search_type, current_point, primal_improved)

        return primal_improved



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.common import deprecated
from pyomo.core import Constraint, value, TransformationFactory
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.repn.standard_repn import generate_standard_repn

logger = logging.getLogger('pyomo.contrib.preprocessing')


@TransformationFactory.register(
    'core.tighten_constraints_from_vars',
    doc="[DEPRECATED] Tightens upper and lower bound on linear constraints.",
)
@deprecated(
    "Use of the constraint tightener transformation is deprecated. "
    "Its functionality may be partially replicated using "
    "`pyomo.contrib.fbbt.compute_bounds_on_expr(constraint.body)`.",
    version='5.7',
)
class TightenConstraintFromVars(IsomorphicTransformation):
    """Tightens upper and lower bound on constraints based on variable bounds.

    Iterates through each variable and tightens the constraint bounds using
    the inferred values from the variable bounds.

    For now, this only operates on linear constraints.

    """

    def __init__(self):
        super(TightenConstraintFromVars, self).__init__()

    def _apply_to(self, model):
        """Apply the transformation."""
        for constr in model.component_data_objects(
            ctype=Constraint, active=True, descend_into=True
        ):
            repn = generate_standard_repn(constr.body)
            if not repn.is_linear():
                continue

            # tighten the constraint bound as much as possible
            LB = UB = 0
            if repn.constant:
                LB = UB = repn.constant

            # loop through each coefficient and variable pair
            for var, coef in zip(repn.linear_vars, repn.linear_coefs):
                # Calculate bounds using interval arithmetic
                if coef >= 0:
                    if var.has_ub():
                        UB = UB + coef * value(var.ub)
                    else:
                        UB = float('Inf')
                    if var.has_lb():
                        LB = LB + coef * value(var.lb)
                    else:
                        LB = float('-Inf')
                else:
                    # coef is negative, so signs switch
                    if var.has_lb():
                        UB = UB + coef * value(var.lb)
                    else:
                        UB = float('Inf')
                    if var.has_ub():
                        LB = LB + coef * value(var.ub)
                    else:
                        LB = float('-Inf')

            # if inferred bound is tighter, replace bound
            new_ub = min(value(constr.upper), UB) if constr.has_ub() else UB
            new_lb = max(value(constr.lower), LB) if constr.has_lb() else LB

            constr.set_value((new_lb, constr.body, new_ub))

            if UB < LB:
                logger.error(
                    "Infeasible variable bounds: "
                    "Constraint %s has inferred LB %s > UB %s"
                    % (constr.name, new_lb, new_ub)
                )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Utility functions and classes for the GDPopt solver."""

from contextlib import contextmanager
import logging
from math import fabs
import sys

from pyomo.common import timing
from pyomo.common.collections import ComponentSet
from pyomo.common.deprecation import deprecation_warning
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from pyomo.contrib.mcpp.pyomo_mcpp import mcpp_available, McCormick
from pyomo.core import (
    Block,
    Constraint,
    minimize,
    Objective,
    Reals,
    Reference,
    TransformationFactory,
    value,
    Var,
)
from pyomo.core.expr.numvalue import native_types
from pyomo.gdp import Disjunct, Disjunction
from pyomo.gdp.util import _parent_disjunct
from pyomo.opt import SolverFactory


class _DoNothing:
    """Do nothing, literally.

    This class is used in situations of "do something if attribute exists."
    """

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        pass

    def __getattr__(self, attr):
        def _do_nothing(*args, **kwargs):
            pass

        return _do_nothing


class SuppressInfeasibleWarning:
    """Suppress the infeasible model warning message from solve().

    The "WARNING: Loading a SolverResults object with a warning status" warning
    message from calling solve() is often unwanted, but there is no clear way
    to suppress it.

    This is modeled on LoggingIntercept from pyomo.common.log,
    but different in function.

    """

    class InfeasibleWarningFilter(logging.Filter):
        def filter(self, record):
            return not record.getMessage().startswith(
                "Loading a SolverResults object with a warning status into model"
            )

    warning_filter = InfeasibleWarningFilter()

    def __enter__(self):
        logger = logging.getLogger('pyomo.core')
        logger.addFilter(self.warning_filter)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        logger = logging.getLogger('pyomo.core')
        logger.removeFilter(self.warning_filter)


def solve_continuous_problem(m, config):
    logger = config.logger
    logger.info('Problem has no discrete decisions.')
    obj = next(m.component_data_objects(Objective, active=True))
    if any(
        c.body.polynomial_degree() not in (1, 0)
        for c in m.component_data_objects(Constraint, active=True, descend_into=Block)
    ) or obj.polynomial_degree() not in (1, 0):
        logger.info(
            "Your model is an NLP (nonlinear program). "
            "Using NLP solver %s to solve." % config.nlp_solver
        )
        results = SolverFactory(config.nlp_solver).solve(m, **config.nlp_solver_args)
        return results
    else:
        logger.info(
            "Your model is an LP (linear program). "
            "Using LP solver %s to solve." % config.mip_solver
        )
        results = SolverFactory(config.mip_solver).solve(m, **config.mip_solver_args)
        return results


def move_nonlinear_objective_to_constraints(util_block, logger):
    m = util_block.parent_block()
    discrete_obj = next(
        m.component_data_objects(Objective, descend_into=True, active=True)
    )
    if discrete_obj.polynomial_degree() in (1, 0):
        # Nothing to move
        return None

    # Move the objective to the constraints if it is nonlinear
    logger.info("Objective is nonlinear. Moving it to constraint set.")

    util_block.objective_value = Var(domain=Reals, initialize=0)
    if mcpp_available():
        mc_obj = McCormick(discrete_obj.expr)
        util_block.objective_value.setub(mc_obj.upper())
        util_block.objective_value.setlb(mc_obj.lower())
    else:
        # Use Pyomo's contrib.fbbt package
        lb, ub = compute_bounds_on_expr(discrete_obj.expr)
        if discrete_obj.sense == minimize:
            util_block.objective_value.setlb(lb)
        else:
            util_block.objective_value.setub(ub)

    if discrete_obj.sense == minimize:
        util_block.objective_constr = Constraint(
            expr=util_block.objective_value >= discrete_obj.expr
        )
    else:
        util_block.objective_constr = Constraint(
            expr=util_block.objective_value <= discrete_obj.expr
        )
    # Deactivate the original objective and add this new one.
    discrete_obj.deactivate()
    util_block.objective = Objective(
        expr=util_block.objective_value, sense=discrete_obj.sense
    )

    # Add the new variable and constraint to the working lists
    util_block.algebraic_variable_list.append(util_block.objective_value)
    if hasattr(util_block, 'constraint_list'):
        util_block.constraint_list.append(util_block.objective_constr)
    # If we moved the objective, return the original in case we want to
    # restore it later
    return discrete_obj


def a_logger(str_or_logger):
    """Returns a logger when passed either a logger name or logger object."""
    if isinstance(str_or_logger, logging.Logger):
        return str_or_logger
    else:
        return logging.getLogger(str_or_logger)


def copy_var_list_values(
    from_list,
    to_list,
    config,
    skip_stale=False,
    skip_fixed=True,
    ignore_integrality=False,
):
    """Copy variable values from one list to another.

    Rounds to Binary/Integer if necessary
    Sets to zero for NonNegativeReals if necessary
    """
    if ignore_integrality:
        deprecation_warning(
            "The 'ignore_integrality' argument no longer has any functionality.",
            version="6.4.2",
        )

    if len(from_list) != len(to_list):
        raise ValueError('The lengths of from_list and to_list do not match.')

    for v_from, v_to in zip(from_list, to_list):
        if skip_stale and v_from.stale:
            continue  # Skip stale variable values.
        if skip_fixed and v_to.is_fixed():
            continue  # Skip fixed variables.
        v_to.set_value(value(v_from, exception=False), skip_validation=True)


def fix_discrete_var(var, val, config):
    """Fixes the discrete variable var to val, rounding to the nearest integer
    or not, depending on if rounding is specified in config and what the integer
    tolerance is."""
    if val is None:
        return
    if var.is_continuous():
        var.set_value(val, skip_validation=True)
    elif fabs(val - round(val)) > config.integer_tolerance:
        raise ValueError(
            "Integer variable '%s' cannot be fixed to value %s because it "
            "is not within the specified integer tolerance of %s."
            % (var.name, val, config.integer_tolerance)
        )
    else:
        # variable is integer and within tolerance
        if config.round_discrete_vars:
            var.fix(int(round(val)))
        else:
            var.fix(val, skip_validation=True)


class fix_discrete_solution_in_subproblem:
    def __init__(
        self,
        true_disjuncts,
        boolean_var_values,
        integer_var_values,
        subprob_util_block,
        config,
        solver,
    ):
        self.True_disjuncts = true_disjuncts
        self.boolean_var_values = boolean_var_values
        self.discrete_var_values = integer_var_values
        self.subprob_util_block = subprob_util_block
        self.config = config

    def __enter__(self):
        # fix subproblem Blocks according to the discrete problem solution
        fixed = []
        for block in self.subprob_util_block.disjunct_list:
            if block in self.True_disjuncts:
                block.binary_indicator_var.fix(1)
                fixed.append(block.name)
            else:
                block.deactivate()
                block.binary_indicator_var.fix(0)
        self.config.logger.debug(
            "Fixed the following Disjuncts to 'True': %s" % ", ".join(fixed)
        )

        fixed_bools = []
        for subprob_bool, val in zip(
            self.subprob_util_block.non_indicator_boolean_variable_list,
            self.boolean_var_values,
        ):
            subprob_binary = subprob_bool.get_associated_binary()
            if val:
                subprob_binary.fix(1)
            else:
                subprob_binary.fix(0)
            fixed_bools.append("%s = %s" % (subprob_bool.name, val))
        self.config.logger.debug(
            "Fixed the following Boolean variables: %s" % ", ".join(fixed_bools)
        )

        # Fix subproblem discrete variables according to the discrete problem
        # solution
        if self.config.force_subproblem_nlp:
            fixed_discrete = []
            for subprob_var, val in zip(
                self.subprob_util_block.discrete_variable_list, self.discrete_var_values
            ):
                fix_discrete_var(subprob_var, val, self.config)
                fixed_discrete.append("%s = %s" % (subprob_var.name, val))
            self.config.logger.debug(
                "Fixed the following integer variables: "
                "%s" % ", ".join(fixed_discrete)
            )

        # Call the subproblem initialization callback
        self.config.subproblem_initialization_method(
            self.True_disjuncts,
            self.boolean_var_values,
            self.discrete_var_values,
            self.subprob_util_block,
        )

        return self

    def __exit__(self, type, value, traceback):
        # unfix all subproblem blocks
        for block in self.subprob_util_block.disjunct_list:
            block.activate()
            block.binary_indicator_var.unfix()

        # unfix all the formerly-Boolean variables
        for bool_var in self.subprob_util_block.non_indicator_boolean_variable_list:
            bool_var.get_associated_binary().unfix()

        # unfix all discrete variables and restore them to their original values
        if self.config.force_subproblem_nlp:
            for subprob_var in self.subprob_util_block.discrete_variable_list:
                subprob_var.fixed = False

        # [ESJ 2/25/22] We don't need to reset the values of the continuous
        # variables because we will initialize them based on the discrete
        # problem solution before we solve again.


class fix_discrete_problem_solution_in_subproblem(fix_discrete_solution_in_subproblem):
    def __init__(self, discrete_prob_util_block, subproblem_util_block, solver, config):
        self.discrete_prob_util_block = discrete_prob_util_block
        self.subprob_util_block = subproblem_util_block
        self.solver = solver
        self.config = config

    def __enter__(self):
        # fix subproblem Blocks according to the discrete problem solution
        fixed = []
        for disjunct, block in zip(
            self.discrete_prob_util_block.disjunct_list,
            self.subprob_util_block.disjunct_list,
        ):
            if not disjunct.indicator_var.value:
                block.deactivate()
                block.binary_indicator_var.fix(0)
            else:
                block.binary_indicator_var.fix(1)
                fixed.append(block.name)
        self.config.logger.debug(
            "Fixed the following Disjuncts to 'True': %s" % ", ".join(fixed)
        )

        fixed_bools = []
        for discrete_problem_bool, subprob_bool in zip(
            self.discrete_prob_util_block.non_indicator_boolean_variable_list,
            self.subprob_util_block.non_indicator_boolean_variable_list,
        ):
            discrete_problem_binary = discrete_problem_bool.get_associated_binary()
            subprob_binary = subprob_bool.get_associated_binary()
            val = discrete_problem_binary.value
            if val is None:
                # If it's None, it's not yet constrained in discrete problem:
                # make an arbitrary decision for now, and store it in the
                # discrete problem so the no-good cut will be right.
                discrete_problem_binary.set_value(1)
                subprob_binary.fix(1)
                bool_val = True
            elif val > 0.5:
                subprob_binary.fix(1)
                bool_val = True
            else:
                subprob_binary.fix(0)
                bool_val = False
            fixed_bools.append("%s = %s" % (subprob_bool.name, bool_val))
        self.config.logger.debug(
            "Fixed the following Boolean variables: %s" % ", ".join(fixed_bools)
        )

        # Fix subproblem discrete variables according to the discrete problem
        # solution
        if self.config.force_subproblem_nlp:
            fixed_discrete = []
            for discrete_problem_var, subprob_var in zip(
                self.discrete_prob_util_block.discrete_variable_list,
                self.subprob_util_block.discrete_variable_list,
            ):
                # [ESJ 1/24/21]: We don't check if discrete problem_var
                # actually has a value here because we are going to have to do
                # that error checking later. This is because the subproblem
                # could have discrete variables that aren't in the discrete
                # problem and vice versa since discrete problem is linearized,
                # but subproblem is a specific realization of the disjuncts. All
                # this means we don't have enough info to do it here.
                fix_discrete_var(subprob_var, discrete_problem_var.value, self.config)
                fixed_discrete.append(
                    "%s = %s" % (subprob_var.name, discrete_problem_var.value)
                )
            self.config.logger.debug(
                "Fixed the following integer variables: "
                "%s" % ", ".join(fixed_discrete)
            )

        # Call the subproblem initialization callback
        self.config.subproblem_initialization_method(
            self.solver, self.subprob_util_block, self.discrete_prob_util_block
        )

        return self


def is_feasible(model, config):
    """Checks to see if the algebraic model is feasible in its current state.

    Checks variable bounds and active constraints. Not for use with
    untransformed GDP models.

    """
    disj = next(model.component_data_objects(ctype=Disjunct, active=True), None)
    if disj is not None:
        raise NotImplementedError(
            "Found active disjunct %s. "
            "This function is not intended to check "
            "feasibility of disjunctive models, "
            "only transformed subproblems." % disj.name
        )

    config.logger.debug('Checking if model is feasible.')
    for constr in model.component_data_objects(
        ctype=Constraint, active=True, descend_into=True
    ):
        # Check constraint lower bound
        if constr.lower is not None and (
            value(constr.lower) - value(constr.body) >= config.constraint_tolerance
        ):
            config.logger.info(
                '%s: body %s < LB %s'
                % (constr.name, value(constr.body), value(constr.lower))
            )
            return False
        # check constraint upper bound
        if constr.upper is not None and (
            value(constr.body) - value(constr.upper) >= config.constraint_tolerance
        ):
            config.logger.info(
                '%s: body %s > UB %s'
                % (constr.name, value(constr.body), value(constr.upper))
            )
            return False
    for var in model.component_data_objects(ctype=Var, descend_into=True):
        # Check variable lower bound
        if var.has_lb() and value(var.lb) - value(var) >= config.variable_tolerance:
            config.logger.info(
                '%s: value %s < LB %s' % (var.name, value(var), value(var.lb))
            )
            return False
        # Check variable upper bound
        if var.has_ub() and value(var) - value(var.ub) >= config.variable_tolerance:
            config.logger.info(
                '%s: value %s > UB %s' % (var.name, value(var), value(var.ub))
            )
            return False
    config.logger.info('Model is feasible.')
    return True


@contextmanager
def time_code(timing_data_obj, code_block_name, is_main_timer=False):
    """Starts timer at entry, stores elapsed time at exit

    If `is_main_timer=True`, the start time is stored in the timing_data_obj,
    allowing calculation of total elapsed time 'on the fly' (e.g. to enforce
    a time limit) using `get_main_elapsed_time(timing_data_obj)`.
    """
    start_time = timing.default_timer()
    if is_main_timer:
        timing_data_obj.main_timer_start_time = start_time
    yield
    elapsed_time = timing.default_timer() - start_time
    prev_time = timing_data_obj.get(code_block_name, 0)
    timing_data_obj[code_block_name] = prev_time + elapsed_time


def get_main_elapsed_time(timing_data_obj):
    """Returns the time since entering the main `time_code` context"""
    current_time = timing.default_timer()
    try:
        return current_time - timing_data_obj.main_timer_start_time
    except AttributeError as e:
        if 'main_timer_start_time' in str(e):
            raise e from AttributeError(
                "You need to be in a 'time_code' context to use "
                "`get_main_elapsed_time()`."
            )


@contextmanager
def lower_logger_level_to(logger, level=None, tee=False):
    """Increases logger verbosity by lowering reporting level."""
    if tee:  # we want pretty stuff
        level = logging.INFO  # we need to be at least this verbose for tee to
        # work
        handlers = [h for h in logger.handlers]
        logger.handlers.clear()
        logger.propagate = False
        # Send logging to stdout
        sh = logging.StreamHandler(sys.stdout)
        # set it to the logger level first, we'll change it below if it needs to
        # become more verbose for tee
        sh.setLevel(logger.getEffectiveLevel())
        sh.setFormatter(logging.Formatter('%(message)s'))
        logger.addHandler(sh)

    level_changed = False
    if level is not None and logger.getEffectiveLevel() > level:
        # If logger level is higher (less verbose), decrease it
        old_logger_level = logger.level
        logger.setLevel(level)
        if tee:
            sh.setLevel(level)
        level_changed = True

    try:
        yield
    finally:
        if tee:
            logger.handlers.clear()
            for h in handlers:
                logger.addHandler(h)
            logger.propagate = True
        if level_changed:
            logger.setLevel(old_logger_level)


def _add_bigm_constraint_to_transformed_model(m, constraint, block):
    """Adds the given constraint to the discrete problem model as if it had
    been on the model originally, before the bigm transformation was called.
    Note this method doesn't actually add the constraint to the model, it just
    takes a constraint that has been added and transforms it.

    Also note that this is not a general method: We know several special
    things in the case of adding OA cuts:
    * No one is going to have a bigm Suffix or arg for this cut--we're
    definitely calculating our own value of M.
    * constraint is for sure a ConstraintData--we don't need to handle anything
    else.
    * We know that we originally called bigm with the default arguments to the
    transformation, so we can assume all of those for this as well. (This is
    part of the reason this *isn't* a general method, what to do about this
    generally is a hard question.)

    Parameters
    ----------
    m: Discrete problem model that has been transformed with bigm.
    constraint: Already-constructed ConstraintData somewhere on m
    block: The block that constraint lives on. This Block may or may not be on
           a Disjunct.
    """
    # Find out it if this constraint really is on a Disjunct. If not, then
    # it's global and we don't actually need to do anything.
    parent_disjunct = block
    if parent_disjunct.ctype is not Disjunct:
        parent_disjunct = _parent_disjunct(block)

    if parent_disjunct is None:
        # the constraint is global, there's nothing to do.
        return

    bigm = TransformationFactory('gdp.bigm')
    # We're fine with default state, but because we're not using apply_to, we
    # need to set it.
    bigm._config = bigm.CONFIG()
    # ESJ: This function doesn't handle ConstraintDatas, and bigm is not
    # sufficiently modular to have a function that does at the moment, so I'm
    # making a Reference to the ComponentData so that it will look like an
    # indexed component for now. If I redesign bigm at some point, then this
    # could be prettier.
    bigm._transform_constraint(
        Reference(constraint),
        parent_disjunct,
        None,
        [],
        [],
        1 - parent_disjunct.binary_indicator_var,
    )
    # Now get rid of it because this is a class attribute!
    del bigm._config
